#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

"""
Cutting plane-based GDP reformulation.

Implements a general cutting plane-based reformulation for linear and
convex GDPs.
"""
from __future__ import division
try:
    from collections import OrderedDict
except:
    from ordereddict import OrderedDict

from pyomo.common.modeling import unique_component_name
from pyomo.core import (
    Any, Block, Constraint, Objective, Param, Var, SortComponents,
    Transformation, TransformationFactory, value, TransformationFactory
)
from pyomo.opt import SolverFactory

from pyomo.gdp import Disjunct, Disjunction, GDP_Error
from pyomo.gdp.util import (
    verify_successful_solve, NORMAL, INFEASIBLE, NONOPTIMAL
)

from six import iterkeys, itervalues

import math
import logging
logger = logging.getLogger('pyomo.gdp.cuttingplane')

# TODO: this should be an option probably, right?
# do I have other options that won't be mad about the quadratic objective in the
# separation problem?
SOLVER = 'ipopt'
stream_solvers = False


@TransformationFactory.register('gdp.cuttingplane', doc="Relaxes a linear disjunctive model by "
          "adding cuts from convex hull to Big-M relaxation.")
class CuttingPlane_Transformation(Transformation):

    def __init__(self):
        super(CuttingPlane_Transformation, self).__init__()

    def _apply_to(self, instance, bigM=None, **kwds):
        options = kwds.pop('options', {})

        if kwds:
            logger.warning("GDP(CuttingPlanes): unrecognized keyword arguments:"
                           "\n%s" % ( '\n'.join(iterkeys(kwds)), ))
        if options:
            logger.warning("GDP(CuttingPlanes): unrecognized options:\n%s"
                        % ( '\n'.join(iterkeys(options)), ))

        instance_rBigM, instance_rHull, var_info, transBlockName \
            = self._setup_subproblems(instance, bigM)

        self._generate_cuttingplanes(
            instance, instance_rBigM, instance_rHull, var_info, transBlockName)


    def _setup_subproblems(self, instance, bigM):
        # create transformation block
        transBlockName, transBlock = self._add_relaxation_block(
            instance,
            '_pyomo_gdp_cuttingplane_relaxation')

        # We store a list of all vars so that we can efficiently
        # generate maps among the subproblems
        transBlock.all_vars = list(v for v in instance.component_data_objects(
            Var,
            descend_into=(Block, Disjunct),
            sort=SortComponents.deterministic) if not v.is_fixed())

        # we'll store all the cuts we add together
        transBlock.cuts = Constraint(Any)

        # get bigM and hull relaxations
        bigMRelaxation = TransformationFactory('gdp.bigm')
        hullRelaxation = TransformationFactory('gdp.hull')
        relaxIntegrality = TransformationFactory('core.relax_integrality')

        #
        # Generalte the Hull relaxation (used for the separation
        # problem to generate cutting planes
        #
        instance_rHull = hullRelaxation.create_using(instance)
        # This relies on relaxIntegrality relaxing variables on deactivated
        # blocks, which should be fine.
        relaxIntegrality.apply_to(instance_rHull)

        #
        # Reformulate the instance using the BigM relaxation (this will
        # be the final instance returned to the user)
        #
        bigMRelaxation.apply_to(instance, bigM=bigM)

        #
        # Generate the continuous relaxation of the BigM transformation
        #
        instance_rBigM = relaxIntegrality.create_using(instance)

        #
        # Add the xstar parameter for the Hull problem
        #
        transBlock_rHull = instance_rHull.component(transBlockName)
        #
        # this will hold the solution to rbigm each time we solve it. We
        # add it to the transformation block so that we don't have to
        # worry about name conflicts.
        transBlock_rHull.xstar = Param(
            range(len(transBlock.all_vars)), mutable=True, default=None)

        transBlock_rBigM = instance_rBigM.component(transBlockName)

        #
        # Generate the mapping between the variables on all the
        # instances and the xstar parameter
        #
        var_info = tuple(
            (v,
             transBlock_rBigM.all_vars[i],
             transBlock_rHull.all_vars[i],
             transBlock_rHull.xstar[i])
            for i,v in enumerate(transBlock.all_vars))

        #
        # Add the separation objective to the hull subproblem
        #
        self._add_separation_objective(var_info, transBlock_rHull)

        return instance_rBigM, instance_rHull, var_info, transBlockName


    def _generate_cuttingplanes(
            self, instance, instance_rBigM, instance_rHull,
            var_info, transBlockName):

        opt = SolverFactory(SOLVER)

        improving = True
        iteration = 0
        prev_obj = float("inf")
        epsilon = 0.01

        transBlock = instance.component(transBlockName)
        transBlock_rBigM = instance_rBigM.component(transBlockName)

        # We try to grab the first active objective. If there is more
        # than one, the writer will yell when we try to solve below. If
        # there are 0, we will yell here.
        rBigM_obj = next(instance_rBigM.component_data_objects(
            Objective, active=True), None)
        if rBigM_obj is None:
            raise GDP_Error("Cannot apply cutting planes transformation "
                            "without an active objective in the model!")

        while (improving):
            # solve rBigM, solution is xstar
            results = opt.solve(instance_rBigM, tee=stream_solvers)
            if verify_successful_solve(results) is not NORMAL:
                logger.warning("GDP.cuttingplane: Relaxed BigM subproblem "
                               "did not solve normally. Stopping cutting "
                               "plane generation.\n\n%s" % (results,))
                return

            rBigM_objVal = value(rBigM_obj)
            logger.warning("gdp.cuttingplane: rBigM objective = %s"
                           % (rBigM_objVal,))

            # copy over xstar
            for x_bigm, x_rbigm, x_hull, x_star in var_info:
                x_star.value = x_rbigm.value
                # initialize the X values
                x_hull.value = x_rbigm.value

            # solve separation problem to get xhat.
            results = opt.solve(instance_rHull, tee=stream_solvers)
            if verify_successful_solve(results) is not NORMAL:
                logger.warning("GDP.cuttingplane: Hull separation subproblem "
                               "did not solve normally. Stopping cutting "
                               "plane generation.\n\n%s" % (results,))
                return

            self._add_cut(var_info, transBlock, transBlock_rBigM)

            # decide whether or not to keep going: check absolute difference
            # close to 0, relative difference further from 0.
            obj_diff = prev_obj - rBigM_objVal
            improving = math.isinf(obj_diff) or \
                        ( abs(obj_diff) > epsilon if abs(rBigM_objVal) < 1 else
                          abs(obj_diff/prev_obj) > epsilon )

            prev_obj = rBigM_objVal
            iteration += 1


    def _add_relaxation_block(self, instance, name):
        # creates transformation block with a unique name based on name, adds it
        # to instance, and returns it.
        transBlockName = unique_component_name(
            instance,
            '_pyomo_gdp_cuttingplane_relaxation')
        transBlock = Block()
        instance.add_component(transBlockName, transBlock)
        return transBlockName, transBlock


    def _add_separation_objective(self, var_info, transBlock_rHull):
        # Deactivate any/all other objectives
        for o in transBlock_rHull.model().component_data_objects(Objective):
            o.deactivate()

        obj_expr = 0
        for x_bigm, x_rbigm, x_hull, x_star in var_info:
            obj_expr += (x_hull - x_star)**2
        # add separation objective to transformation block
        transBlock_rHull.separation_objective = Objective(expr=obj_expr)


    def _add_cut(self, var_info, transBlock, transBlock_rBigM):
        # add cut to BM and rBM
        cut_number = len(transBlock.cuts)
        logger.warning("gdp.cuttingplane: Adding cut %s to BM model."
                       % (cut_number,))

        cutexpr_bigm = 0
        cutexpr_rBigM = 0
        for x_bigm, x_rbigm, x_hull, x_star in var_info:
            # xhat = x_hull.value
            cutexpr_bigm += (
                x_hull.value - x_star.value)*(x_bigm - x_hull.value)
            cutexpr_rBigM += (
                x_hull.value - x_star.value)*(x_rbigm - x_hull.value)

        transBlock.cuts.add(cut_number, cutexpr_bigm >= 0)
        transBlock_rBigM.cuts.add(cut_number, cutexpr_rBigM >= 0)



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
#
# Unit Tests for pyomo.opt.problem.ampl
#

import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep

import pyutilib.th as unittest
import pyutilib.services
import pyutilib.common

import pyomo.opt

old_tempdir = pyutilib.services.TempfileManager.tempdir

def filter(text):
    return 'Problem:' in text or text.startswith('NAME')

def filter_nl(text):
    return '# problem'

solver = None
class Test(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        global solvers
        import pyomo.environ
        solvers = pyomo.opt.check_available_solvers('glpk')

    def setUp(self):
        pyutilib.services.TempfileManager.tempdir = currdir

    def tearDown(self):
        pyutilib.services.TempfileManager.clear_tempfiles()
        pyutilib.services.TempfileManager.tempdir = old_tempdir

    def test3_write_nl(self):
        """ Convert from AMPL to NL """
        self.model = pyomo.opt.AmplModel(currdir+'test3.mod')
        """ Convert from MOD+DAT to NL """
        try:
            self.model.write(currdir+'test3.nl')
        except pyutilib.common.ApplicationError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail("Unexpected ApplicationError - ampl is enabled "
                          "but not available: '%s'" % str(err))
            return
        except pyomo.opt.ConverterError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail("Unexpected ConverterError - ampl is enabled "
                          "but not available: '%s'" % str(err))
            return
        self.assertFileEqualsBaseline(currdir+'test3.nl', currdir+'test3.baseline.nl', filter=filter_nl, tolerance=1e-6)

    def test3_write_lp(self):
        """ Convert from AMPL to LP """
        self.model = pyomo.opt.AmplModel(currdir+'test3.mod')
        try:
            self.model.write(currdir+'test3.lp')
        except pyutilib.common.ApplicationError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("glpsol"):
                self.fail("Unexpected ApplicationError - glpsol is enabled "
                          "but not available: '%s'" % str(err))
            return
        except pyomo.opt.ConverterError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("glpsol"):
                self.fail("Unexpected ConverterError - glpsol is enabled "
                          "but not available: '%s'" % str(err))
            return
        self.assertFileEqualsBaseline(currdir+'test3.lp', currdir+'test3.baseline.lp', filter=filter, tolerance=1e-6)

    def test3_write_mps(self):
        """ Convert from AMPL to MPS """
        if not pyomo.common.Executable("ampl"):
            self.skipTest("The ampl executable is not available")
        self.model = pyomo.opt.AmplModel(currdir+'test3.mod')
        try:
            self.model.write(currdir+'test3.mps')
        except pyutilib.common.ApplicationError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail("Unexpected ApplicationError - ampl is enabled "
                          "but not available: '%s'" % str(err))
            return
        except pyomo.opt.ConverterError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail("Unexpected ConverterError - ampl is enabled "
                          "but not available: '%s'" % str(err))
            return
        self.assertFileEqualsBaseline(currdir+'test3.mps', currdir+'test3.baseline.mps', filter=filter, tolerance=1e-6)

    def test3a_write_nl(self):
        """ Convert from AMPL to NL """
        self.model = pyomo.opt.AmplModel(currdir+'test3a.mod', currdir+'test3a.dat')
        try:
            self.model.write(currdir+'test3a.nl')
        except pyutilib.common.ApplicationError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail("Unexpected ApplicationError - ampl is enabled "
                          "but not available: '%s'" % str(err))
            return
        except pyomo.opt.ConverterError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail("Unexpected ConverterError - ampl is enabled "
                          "but not available: '%s'" % str(err))
            return
        self.assertFileEqualsBaseline(currdir+'test3a.nl', currdir+'test3.baseline.nl', filter=filter_nl, tolerance=1e-6)

    def test3a_write_lp(self):
        """ Convert from AMPL to LP """
        self.model = pyomo.opt.AmplModel(currdir+'test3a.mod', currdir+'test3a.dat')
        try:
            self.model.write(currdir+'test3a.lp')
        except pyutilib.common.ApplicationError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("glpsol"):
                self.fail("Unexpected ApplicationError - glpsol is enabled "
                          "but not available: '%s'" % str(err))
            return
        except pyomo.opt.ConverterError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("glpsol"):
                self.fail("Unexpected ConverterError - glpsol is enabled "
                          "but not available: '%s'" % str(err))
            return
        self.assertFileEqualsBaseline(currdir+'test3a.lp', currdir+'test3.baseline.lp', filter=filter, tolerance=1e-6)

    def test3a_write_mps(self):
        """ Convert from AMPL to MPS """
        if not pyomo.common.Executable("ampl"):
            self.skipTest("The ampl executable is not available")
        self.model = pyomo.opt.AmplModel(currdir+'test3a.mod', currdir+'test3a.dat')
        try:
            self.model.write(currdir+'test3a.mps')
        except pyutilib.common.ApplicationError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail("Unexpected ApplicationError - ampl is enabled "
                          "but not available: '%s'" % str(err))
            return
        except pyomo.opt.ConverterError:
            err = sys.exc_info()[1]
            if pyomo.common.Executable("ampl"):
                self.fail("Unexpected ConverterError - ampl is enabled "
                          "but not available: '%s'" % str(err))
            return
        self.assertFileEqualsBaseline(currdir+'test3a.mps', currdir+'test3.baseline.mps', filter=filter, tolerance=1e-6)

    def test3_solve(self):
        if not 'glpk' in solvers:
            self.skipTest("glpk solver is not available")
        self.model = pyomo.opt.AmplModel(currdir+'test3.mod')
        opt = pyomo.opt.SolverFactory('glpk')
        results = opt.solve(self.model, keepfiles=False)
        results.write(filename=currdir+'test3.out', format='json')
        self.assertMatchesJsonBaseline(currdir+'test3.out', currdir+'test3.baseline.out', tolerance=1e-6)

    def test3a_solve(self):
        if not 'glpk' in solvers:
            self.skipTest("glpk solver is not available")
        self.model = pyomo.opt.AmplModel(currdir+'test3a.mod', currdir+'test3a.dat')
        opt = pyomo.opt.SolverFactory('glpk')
        results = opt.solve(self.model, keepfiles=False)
        results.write(filename=currdir+'test3a.out', format='json')
        self.assertMatchesJsonBaseline(currdir+'test3a.out', currdir+'test3.baseline.out', tolerance=1e-6)


if __name__ == "__main__":
    unittest.main()




#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
#
# Test the Pyomo command-line interface
#

import re
import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep

import pyutilib.subprocess
import pyutilib.th as unittest
from pyutilib.misc import setup_redirect, reset_redirect

from pyomo.common.dependencies import yaml_available
import pyomo.core
import pyomo.scripting.pyomo_main as main
from pyomo.opt import check_available_solvers

from six import StringIO

if os.path.exists(sys.exec_prefix+os.sep+'bin'+os.sep+'coverage'):
    executable=sys.exec_prefix+os.sep+'bin'+os.sep+'coverage -x '
else:
    executable=sys.executable

def filter_fn(line):
    tmp = line.strip()
    return tmp.startswith('Disjunct') or tmp.startswith('DEPRECATION') or tmp.startswith('DiffSet') or line.startswith('    ') or tmp.startswith("Differential") or tmp.startswith("DerivativeVar") or tmp.startswith("InputVar") or tmp.startswith('StateVar') or tmp.startswith('Complementarity')


_diff_tol = 1e-6

solvers = None
class BaseTester(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        global solvers
        import pyomo.environ
        solvers = check_available_solvers('glpk')

    def pyomo(self, cmd, **kwds):
        if 'root' in kwds:
            OUTPUT=kwds['root']+'.out'
            results=kwds['root']+'.jsn'
            self.ofile = OUTPUT
        else:
            OUTPUT=StringIO()
            results='results.jsn'
        setup_redirect(OUTPUT)
        os.chdir(currdir)
        if type(cmd) is list:
            output = main.main(['solve', '--solver=glpk', '--results-format=json', '--save-results=%s' % results] + cmd)
        elif cmd.endswith('json') or cmd.endswith('yaml'):
            output = main.main(['solve', '--results-format=json', '--save-results=%s' % results] + [cmd])
        else:
            args=re.split('[ ]+',cmd)
            output = main.main(['solve', '--solver=glpk', '--results-format=json', '--save-results=%s' % results] + list(args))
        reset_redirect()
        if not 'root' in kwds:
            return OUTPUT.getvalue()
        return output

    def setUp(self):
        self.ofile = None
        if not 'glpk' in solvers:
            self.skipTest("GLPK is not installed")

    def tearDown(self):
        return
        if self.ofile and os.path.exists(self.ofile):
            return
            os.remove(self.ofile)
        if os.path.exists(currdir+'results.jsn'):
            return
            os.remove(currdir+'results.jsn')

    def run_pyomo(self, cmd, root=None):
        cmd = 'pyomo solve --solver=glpk --results-format=json ' \
              '--save-results=%s.jsn %s' % (root, cmd)
        return pyutilib.subprocess.run(cmd, outfile=root+'.out')


class TestJson(BaseTester):

    def test1_simple_pyomo_execution(self):
        # Simple execution of 'pyomo'
        self.pyomo([currdir+'pmedian.py',currdir+'pmedian.dat'], root=currdir+'test1')
        self.assertMatchesJsonBaseline(currdir+"test1.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test1.out')

    def test1a_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' in a subprocess
        self.run_pyomo('%s/pmedian.py %s/pmedian.dat' % (currdir,currdir),
                       root=currdir+'test1a')
        self.assertMatchesJsonBaseline(currdir+"test1a.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test1a.out')

    def test1b_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' with a configuration file
        self.pyomo(currdir+'test1b.json', root=currdir+'test1')
        self.assertMatchesJsonBaseline(currdir+"test1.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test1.out')

    def test2_bad_model_name(self):
        # Run pyomo with bad --model-name option value
        self.pyomo('--model-name=dummy pmedian.py pmedian.dat', root=currdir+'test2')
        def filter2(line):
            return line.startswith('[') or line.startswith('DEPRECATION')
        self.assertFileEqualsBaseline(currdir+"test2.out", currdir+"test2.txt", filter=filter2)

    def test2b_bad_model_name(self):
        # Run pyomo with bad --model-name option value (configfile)
        self.pyomo(currdir+'test2b.json', root=currdir+'test2')
        def filter2(line):
            return line.startswith('[') or line.startswith('DEPRECATION')
        self.assertFileEqualsBaseline(currdir+"test2.out", currdir+"test2.txt", filter=filter2)

    def test3_missing_model_object(self):
        # Run pyomo with model that does not define model object
        self.pyomo('pmedian1.py pmedian.dat', root=currdir+'test3')
        def filter3(line):
            return line.startswith('[') or line.startswith('DEPRECATION')
        self.assertMatchesJsonBaseline(currdir+"test3.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test3.out')

    def test4_valid_modelname_option(self):
        # Run pyomo with good --model-name option value
        self.pyomo('--model-name=MODEL '+currdir+'pmedian1.py pmedian.dat', root=currdir+'test4')
        self.assertMatchesJsonBaseline(currdir+"test4.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test4.out')

    def test4b_valid_modelname_option(self):
        # Run pyomo with good 'object name' option value (configfile)
        self.pyomo(currdir+'test4b.json', root=currdir+'test4b')
        self.assertMatchesJsonBaseline(currdir+"test4b.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test4b.out')

    def test5_create_model_fcn(self):
        #"""Run pyomo with create_model function"""
        self.pyomo('pmedian2.py pmedian.dat', root=currdir+'test5')
        def filter5(line):
            return ("Writing model " in line) or ("Solver results file" in line) or \
                   line.startswith('[') or line.startswith('DEPRECATION')
        self.assertFileEqualsBaseline(currdir+"test5.out", currdir+"test5.txt",filter=filter5,tolerance=_diff_tol)
        os.remove(currdir+'test5.jsn')

    def test5b_create_model_fcn(self):
        # Run pyomo with create_model function (configfile)
        self.pyomo(currdir+'test5b.json', root=currdir+'test5')
        def filter5(line):
            return ("Writing model " in line) or ("Solver results file" in line) or \
                   line.startswith('[') or line.startswith('DEPRECATION')
        self.assertFileEqualsBaseline(currdir+"test5.out", currdir+"test5.txt",filter=filter5,tolerance=_diff_tol)
        os.remove(currdir+'test5.jsn')

    def test8_instanceonly_option(self):
        #"""Run pyomo with --instance-only option"""
        output = self.pyomo('--instance-only pmedian.py pmedian.dat', root=currdir+'test8')
        self.assertEqual(type(output.retval.instance), pyomo.core.ConcreteModel)
        # Check that the results file was NOT created
        self.assertRaises(OSError, lambda: os.remove(currdir+'test8.jsn'))
        os.remove(currdir+'test8.out')

    def test8b_instanceonly_option(self):
        # Run pyomo with --instance-only option (configfile)
        output = self.pyomo(currdir+'test8b.json', root=currdir+'test8')
        self.assertEqual(type(output.retval.instance), pyomo.core.ConcreteModel)
        # Check that the results file was NOT created
        self.assertRaises(OSError, lambda: os.remove(currdir+'test8.jsn'))
        os.remove(currdir+'test8.out')

    def test9_disablegc_option(self):
        #"""Run pyomo with --disable-gc option"""
        output = self.pyomo('--disable-gc pmedian.py pmedian.dat', root=currdir+'test9')
        self.assertEqual(type(output.retval.instance), pyomo.core.ConcreteModel)
        os.remove(currdir+'test9.jsn')
        os.remove(currdir+'test9.out')

    def test9b_disablegc_option(self):
        # Run pyomo with --disable-gc option (configfile)
        output = self.pyomo(currdir+'test9b.json', root=currdir+'test9')
        self.assertEqual(type(output.retval.instance), pyomo.core.ConcreteModel)
        os.remove(currdir+'test9.jsn')
        os.remove(currdir+'test9.out')

    def test12_output_option(self):
        #"""Run pyomo with --output option"""
        self.pyomo('--logfile=%s pmedian.py pmedian.dat' % (currdir+'test12.log'), root=currdir+'test12')
        self.assertMatchesJsonBaseline(currdir+"test12.jsn", currdir+"test12.txt",tolerance=_diff_tol)
        os.remove(currdir+'test12.log')
        os.remove(currdir+'test12.out')

    def test12b_output_option(self):
        # Run pyomo with --output option (configfile)
        self.pyomo(currdir+'test12b.json', root=currdir+'test12')
        self.assertMatchesJsonBaseline(currdir+"test12.jsn", currdir+"test12.txt",tolerance=_diff_tol)
        os.remove('test12b.log')
        os.remove(currdir+'test12.out')

    def test14_concrete_model_with_constraintlist(self):
        # Simple execution of 'pyomo' with a concrete model and constraint lists
        self.pyomo('pmedian4.py', root=currdir+'test14')
        self.assertMatchesJsonBaseline(currdir+"test14.jsn", currdir+"test14.txt",tolerance=_diff_tol)
        os.remove(currdir+'test14.out')

    def test14b_concrete_model_with_constraintlist(self):
        # Simple execution of 'pyomo' with a concrete model and constraint lists (configfile)
        self.pyomo('pmedian4.py', root=currdir+'test14')
        self.assertMatchesJsonBaseline(currdir+"test14.jsn", currdir+"test14.txt",tolerance=_diff_tol)
        os.remove(currdir+'test14.out')

    def test15_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' with options
        self.pyomo(['--solver-options="mipgap=0.02 cuts="', currdir+'pmedian.py', 'pmedian.dat'], root=currdir+'test15')
        self.assertMatchesJsonBaseline(currdir+"test15.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test15.out')

    def test15b_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' with options
        self.pyomo(currdir+'test15b.json', root=currdir+'test15b')
        self.assertMatchesJsonBaseline(currdir+"test15b.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test15b.out')

    def test15c_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' with options
        self.pyomo(currdir+'test15c.json', root=currdir+'test15c')
        self.assertMatchesJsonBaseline(currdir+"test15c.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test15c.out')


@unittest.skipIf(not yaml_available, "YAML not available available")
class TestWithYaml(BaseTester):

    def test15b_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' with options
        self.pyomo(currdir+'test15b.yaml', root=currdir+'test15b')
        self.assertMatchesJsonBaseline(currdir+"test15b.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test15b.out')

    def test15c_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' with options
        self.pyomo(currdir+'test15c.yaml', root=currdir+'test15c')
        self.assertMatchesJsonBaseline(currdir+"test15c.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test15c.out')


if __name__ == "__main__":
    unittest.main()



# Provide some test for rapper
# Author: David L. Woodruff (Sept 2018)

import pyutilib.th as unittest
import tempfile
import sys
import os
import shutil
import json
import pyomo.environ as pyo
import pyomo.pysp.util.rapper as rapper
from pyomo.pysp.scenariotree.tree_structure_model import CreateAbstractScenarioTreeModel
import pyomo as pyomoroot

__author__ = 'David L. Woodruff <DLWoodruff@UCDavis.edu>'
__version__ = 1.5

solvername = "ipopt" # could use almost any solver
solver_available = pyo.SolverFactory(solvername).available(False)

class Test_abstract_rapper(unittest.TestCase):
    """ Test the rapper code."""

    def setUp(self):
        """ Get ready for tests"""

        p = str(pyomoroot.__path__)
        l = p.find("'")
        r = p.find("'", l+1)
        pyomorootpath = p[l+1:r]
        farmpath = pyomorootpath + os.sep + ".." + os.sep + "examples" + \
                   os.sep + "pysp" + os.sep + "farmer"
        farmpath = os.path.abspath(farmpath)
        
        # for AbstractModels
        self.farmer_ReferencePath = farmpath + os.sep + \
                                    "models" + os.sep + "ReferenceModel.py"
        self.farmer_scenarioPath = farmpath + os.sep + \
                                    "scenariodata"
        
    def tearDown(self):
        # from GH: This step is key, as Python keys off the name of the module, not the location.
        #       So, different reference models in different directories won't be detected.
        #       If you don't do this, the symptom is a model that doesn't have the attributes
        #       that the data file expects.
        if "ReferenceModel" in sys.modules:
            del sys.modules["ReferenceModel"]



    def test_Abstract_Construction(self):
        """ see if we can create the solver object for an AbstractModel"""
             
        stsolver = rapper.StochSolver(self.farmer_ReferencePath,
                                      fsfct = None,
                                      tree_model = self.farmer_scenarioPath,
                                      phopts = None)

    @unittest.skipIf(not solver_available,
                     "%s solver is not available" % (solvername,))
    def test_Abstract_ef(self):
        """ see if we can create the solver object for an AbstractModel"""
             
        stsolver = rapper.StochSolver(self.farmer_ReferencePath,
                                      fsfct = None,
                                      tree_model = self.farmer_scenarioPath,
                                      phopts = None)
        ef_sol = stsolver.solve_ef(solvername)
        assert(ef_sol.solver.termination_condition \
               ==  pyo.TerminationCondition.optimal)

# see also foo.py        
if __name__ == '__main__':
    unittest.main()



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________


import re
import os

from six import iteritems

import pyomo.common
import pyutilib.common
import pyutilib.common
import pyutilib.misc

from pyomo.opt.base import *
from pyomo.opt.base.solvers import _extract_version
from pyomo.opt.results import *
from pyomo.opt.solver import *
from pyomo.solvers.mockmip import MockMIP

import logging
logger = logging.getLogger('pyomo.solvers')


@SolverFactory.register('pico', doc='The PICO LP/MIP solver')
class PICO(OptSolver):
    """The PICO LP/MIP solver
    """

    def __new__(cls, *args, **kwds):
        try:
            mode = kwds['solver_io']
            if mode is None:
                mode = 'lp'
            del kwds['solver_io']
        except KeyError:
            mode = 'lp'
        #
        if mode  == 'lp':
            opt = SolverFactory('_pico_shell', **kwds)
            opt.set_problem_format(ProblemFormat.cpxlp)
            return opt
        # PICO's MPS parser seems too buggy to expose
        # this option
#        if mode  == 'mps':
#            opt = SolverFactory('_pico_shell', **kwds)
#            opt.set_problem_format(ProblemFormat.mps)
#            return opt
        #
        if mode == 'nl':
            # PICO does not accept all asl style command line options
            # (-s in particular, which is required for streaming output of
            # all asl solvers). Therefore we need to send it through the cbc_shell
            # instead of ASL
            opt = SolverFactory('_pico_shell',**kwds)
            opt.set_problem_format(ProblemFormat.nl)
        elif mode == 'os':
            opt = SolverFactory('_ossolver', **kwds)
        else:
            logger.error('Unknown IO type: %s' % mode)
            return
        opt.set_options('solver=PICO')
        return opt


@SolverFactory.register('_pico_shell', doc='Shell interface to the PICO MIP solver')
class PICOSHELL(SystemCallSolver):
    """Shell interface to the PICO LP/MIP solver
    """

    def __init__(self, **kwds):
        #
        # Call base constructor
        #
        kwds["type"] = "pico"
        SystemCallSolver.__init__(self, **kwds)
        #
        # Setup valid problem formats, and valid results for each problem format
        #
        self._valid_problem_formats=[ProblemFormat.cpxlp, ProblemFormat.nl, ProblemFormat.mps]
        self._valid_result_formats = {}
        self._valid_result_formats[ProblemFormat.cpxlp] = [ResultsFormat.soln]
        self._valid_result_formats[ProblemFormat.nl] = [ResultsFormat.sol]
        self._valid_result_formats[ProblemFormat.mps] = [ResultsFormat.soln]
        self.set_problem_format(ProblemFormat.cpxlp)

        # Note: Undefined capabilities default to 'None'
        self._capabilities = pyutilib.misc.Options()
        self._capabilities.linear = True
        self._capabilities.integer = True
        #self._capabilities.sos1 = True
        #self._capabilities.sos2 = True

    #
    # Override the base class method so we can update
    #  _default_variable_value
    #
    def set_problem_format(self, format):
        #
        # The NL file interface for PICO does not return sparse
        # results
        #
        if format == ProblemFormat.nl:
            self._default_variable_value = None
        else:
            self._default_variable_value = 0.0
        super(PICOSHELL, self).set_problem_format(format)

    def _default_results_format(self, prob_format):
        if prob_format == ProblemFormat.nl:
            return ResultsFormat.sol
        return ResultsFormat.soln

    def _default_executable(self):
        executable = pyomo.common.Executable("PICO_deprecated_not_supported")
        if not executable:
            logger.warning("Could not locate the 'PICO' executable, "
                           "which is required for solver %s" % self.name)
            self.enable = False
            return None
        return executable.path()

    def _get_version(self):
        """
        Returns a tuple describing the solver executable version.
        """
        solver_exec = self.executable()
        if solver_exec is None:
            return _extract_version('')
        results = pyutilib.subprocess.run([solver_exec, "--version"], timelimit=1)
        # 'PICO --version' seems to print 'pebble <version>, PICO <version>
        # we don't wan't the pebble version being advertised so we split
        # the string at the comma before extracting a version number. It
        # also exits with a nonzero return code so don't bother checking it.
        return _extract_version(results[1].split(',')[1])

    # Nothing needs to be done here
    #def _presolve(self, *args, **kwds):
    #    # let the base class handle any remaining keywords/actions.
    #    SystemCallSolver._presolve(self, *args, **kwds)

    def create_command_line(self,executable,problem_files):
        #
        # Define log file
        #
        if self._log_file is None:
            self._log_file = pyutilib.services.TempfileManager.\
                            create_tempfile(suffix="PICO.log")

        problem_filename_prefix = problem_files[0]
        if '.' in problem_filename_prefix:
            tmp = problem_filename_prefix.split('.')
            if len(tmp) > 2:
                problem_filename_prefix = '.'.join(tmp[:-1])
            else:
                problem_filename_prefix = tmp[0]
        if self._results_format is ResultsFormat.sol:
            self._soln_file = problem_filename_prefix+".sol"
        else:
            self._soln_file = problem_filename_prefix+".soln"

        #
        # Define results file (if the sol external parser is used)
        #
        if self._results_format is ResultsFormat.sol:
            self._results_file = self._soln_file

        #
        # Eventually, these formats will be added to PICO...
        #
        #elif self._results_format == ResultsFormat.osrl:
            #self._results_file = self.tmpDir+os.sep+"PICO.osrl.xml"

        def _check_and_escape_options(options):
            for key, val in iteritems(self.options):
                tmp_k = str(key)
                _bad = ' ' in tmp_k

                tmp_v = str(val)
                if ' ' in tmp_v:
                    if '"' in tmp_v:
                        if "'" in tmp_v:
                            _bad = True
                        else:
                            tmp_v = "'" + tmp_v + "'"
                    else:
                        tmp_v = '"' + tmp_v + '"'

                if _bad:
                    raise ValueError("Unable to properly escape solver option:"
                                     "\n\t%s=%s" % (key, val) )
                yield (tmp_k, tmp_v)

        #
        # Define command line
        #
        cmd = [ executable ]
        if self._timer:
            cmd.insert(0, self._timer)

        if (self.options.mipgap is not None):
            raise ValueError("The mipgap parameter is currently not being "
                             "processed by PICO solver plugin")
        env=os.environ.copy()
        if self._problem_format is None \
                or self._problem_format == ProblemFormat.nl:
            cmd.append(problem_files[0])
            cmd.append('-AMPL')

            opts = ['allowInfiniteIntegerVarBounds=true']
            if "debug" not in self.options:
                opts.append("debug=2")
            for key, val in _check_and_escape_options(self.options):
                if key == 'solver':
                    continue
                opts.append("%s=%s" % ( key, tmp ))
            opts.append('output='+self._soln_file)
            env["PICO_options"] = " ".join(opts)

        elif self._problem_format == ProblemFormat.cpxlp \
                or self._problem_format == ProblemFormat.mps:
            cmd.append('--allowInfiniteIntegerVarBounds=true')
            # This option should appear in the next PICO release,
            # assuming that ever happens
            if self.version() > (1,3,1,0):
                cmd.append('--reportDenseSolution=true')
            if "debug" not in self.options:
                cmd.extend(["--debug", "2"])
            for key, val in _check_and_escape_options(self.options):
                if key == 'mipgap':
                    continue
                cmd.extend(['--'+key, val])
            cmd.extend(['--output', self._soln_file,
                        problem_files[0]])

        return pyutilib.misc.Bunch(cmd=cmd, log_file=self._log_file, env=env)

    def process_logfile(self):
        """
        Process a logfile
        """
        results = SolverResults()

        #
        # Initial values
        #
        #results.solver.statistics.branch_and_bound.number_of_created_subproblems=0
        #results.solver.statistics.branch_and_bound.number_of_bounded_subproblems=0
        soln = Solution()
        soln.objective['__default_objective__'] = {'Value': None}
        #
        # Process logfile
        #
        OUTPUT = open(self._log_file)
        output = "".join(OUTPUT.readlines())
        OUTPUT.close()
        #
        # Parse logfile lines
        #
        for line in output.split("\n"):
            tokens = re.split('[ \t]+',line.strip())
            if len(tokens) > 3 and tokens[0] == "ABORTED:":
                results.solver.status=SolverStatus.aborted
            elif len(tokens) > 1 and tokens[0].startswith("ERROR"):
                results.solver.status=SolverStatus.error
            elif len(tokens) == 3 and tokens[0] == 'Problem' and tokens[2].startswith('infeasible'):
                results.solver.termination_condition = TerminationCondition.infeasible
            elif len(tokens) == 2 and tokens[0] == 'Integer' and tokens[1] == 'Infeasible':
                results.solver.termination_condition = TerminationCondition.infeasible
            elif len(tokens) == 5 and tokens[0] == "Final" and tokens[1] == "Solution:":
                soln.objective['__default_objective__']['Value'] = eval(tokens[4])
                soln.status = SolutionStatus.optimal
            elif len(tokens) == 3 and tokens[0] == "LP" and tokens[1] == "value=":
                soln.objective['__default_objective__']['Value'] = eval(tokens[2])
                soln.status=SolutionStatus.optimal
                if results.problem.sense == ProblemSense.minimize:
                    results.problem.lower_bound = eval(tokens[2])
                else:
                    results.problem.upper_bound = eval(tokens[2])
            elif len(tokens) == 2 and tokens[0] == "Bound:":
                if results.problem.sense == ProblemSense.minimize:
                    results.problem.lower_bound = eval(tokens[1])
                else:
                    results.problem.upper_bound = eval(tokens[1])
            elif len(tokens) == 3 and tokens[0] == "Created":
                results.solver.statistics.branch_and_bound.number_of_created_subproblems = eval(tokens[1])
            elif len(tokens) == 3 and tokens[0] == "Bounded":
                results.solver.statistics.branch_and_bound.number_of_bounded_subproblems = eval(tokens[1])
            elif len(tokens) == 2 and tokens[0] == "sys":
                results.solver.system_time=eval(tokens[1])
            elif len(tokens) == 2 and tokens[0] == "user":
                results.solver.user_time=eval(tokens[1])
            elif len(tokens) == 3 and tokens[0] == "Solving" and tokens[1] == "problem:":
                results.problem.name = tokens[2]
            elif len(tokens) == 4 and tokens[2] == "constraints:":
                results.problem.number_of_constraints = eval(tokens[3])
            elif len(tokens) == 4 and tokens[2] == "variables:":
                results.problem.number_of_variables = eval(tokens[3])
            elif len(tokens) == 4 and tokens[2] == "nonzeros:":
                results.problem.number_of_nonzeros = eval(tokens[3])
            elif len(tokens) == 3 and tokens[1] == "Sense:":
                if tokens[2] == "minimization":
                    results.problem.sense = ProblemSense.minimize
                else:
                    results.problem.sense = ProblemSense.maximize

        if results.solver.status is SolverStatus.aborted:
            soln.optimality=SolutionStatus.unsure
        if soln.status is SolutionStatus.optimal:
            soln.gap=0.0
            results.problem.lower_bound = soln.objective['__default_objective__']['Value']
            results.problem.upper_bound = soln.objective['__default_objective__']['Value']

        if soln.status == SolutionStatus.optimal:
            results.solver.termination_condition = TerminationCondition.optimal

        if not results.solver.status is SolverStatus.error and \
            results.solver.termination_condition in [TerminationCondition.unknown,
                        #TerminationCondition.maxIterations,
                        #TerminationCondition.minFunctionValue,
                        #TerminationCondition.minStepLength,
                        TerminationCondition.globallyOptimal,
                        TerminationCondition.locallyOptimal,
                        TerminationCondition.optimal,
                        #TerminationCondition.maxEvaluations,
                        TerminationCondition.other]:
                results.solution.insert(soln)
        return results


    def process_soln_file(self,results):

        if self._results_format is ResultsFormat.sol:
            return

        # the only suffixes that we extract from PICO are
        # constraint duals. scan through the solver suffix
        # list and throw an exception if the user has
        # specified any others.
        extract_duals = False
        for suffix in self._suffixes:
            if re.match(suffix,"dual"):
                extract_duals = True
            else:
                raise RuntimeError("***PICO solver plugin cannot extract solution suffix="+suffix)

        #if os.path.exists(self.sol_file):
            #results_reader = ReaderFactory(ResultsFormat.sol)
            #results = results_reader(self.sol_file, results, results.solution(0))
            #return

        if not os.path.exists(self._soln_file):
            return
        if len(results.solution) == 0:
            return
        soln = results.solution(0)
        results.problem.num_objectives=1
        tmp=[]
        flag=False
        INPUT = open(self._soln_file, "r")
        lp_flag=None
        var_flag=True
        for line in INPUT:
            tokens = re.split('[ \t]+',line.strip())
            if len(tokens) == 0 or (len(tokens) == 1 and tokens[0]==''):
                continue
            if tokens[0] == "Objective":
                continue
            if lp_flag is None:
                lp_flag = (tokens[0] == "LP")
                continue
            if tokens[0] == "Dual" and tokens[1] == "solution:":
                var_flag=False
                # It looks like we've just been processing primal
                # variables.
                for (var,val) in tmp:
                    if var == 'ONE_VAR_CONSTANT':
                        continue
                    soln.variable[var] = {"Value" : val}
                tmp=[]
                continue
            if len(tokens) < 3:
                print("ERROR", line,tokens)
            tmp.append( (tokens[0],eval(tokens[2])) )
        if var_flag:
            for (var,val) in tmp:
                if var == 'ONE_VAR_CONSTANT':
                    continue
                soln.variable[var] = {"Value" : val}
        else:
            range_duals = {}
            soln_constraints = soln.constraint
            if (lp_flag is True) and (extract_duals is True):
                for (var,val) in tmp:
                    if var.startswith('c_'):
                        soln_constraints[var] = {"Dual" : val}
                    elif var.startswith('r_l_'):
                        range_duals.setdefault(var[4:],[0,0])[0] = val
                    elif var.startswith('r_u_'):
                        range_duals.setdefault(var[4:],[0,0])[1] = val
            # For the range constraints, supply only the dual with the largest
            # magnitude (at least one should always be numerically zero)
            for key,(ld,ud) in iteritems(range_duals):
                if abs(ld) > abs(ud):
                    soln_constraints['r_l_'+key] = {"Dual" : ld}
                else:
                    soln_constraints['r_l_'+key] = {"Dual" : ud}        # Use the same key
        INPUT.close()


@SolverFactory.register('_mock_pico')
class MockPICO(PICOSHELL,MockMIP):
    """A Mock PICO solver used for testing
    """

    def __init__(self, **kwds):
        try:
            PICOSHELL.__init__(self,**kwds)
        except pyutilib.common.ApplicationError: #pragma:nocover
            pass                        #pragma:nocover
        MockMIP.__init__(self,"pico")

    def available(self, exception_flag=True):
        return PICOSHELL.available(self,exception_flag)

    def create_command_line(self,executable,problem_files):
        command = PICOSHELL.create_command_line(self,executable,problem_files)
        MockMIP.create_command_line(self,executable,problem_files)
        return command

    def _default_executable(self):
        return MockMIP.executable(self)

    def _execute_command(self,cmd):
        return MockMIP._execute_command(self,cmd)



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

# the intent of this module is to provide functions to interface from
# a PH client to a set of PH solver servers.

import time
import itertools
import enum

from pyomo.core import *

from six import iteritems, itervalues


class InvocationType(str, enum.Enum):
    SingleInvocation =             'SingleInvocation'
    PerBundleInvocation =          'PerBundleInvocation'
    PerBundleChainedInvocation =   'PerBundleChainedInvocation'
    PerScenarioInvocation =        'PerScenarioInvocation'
    PerScenarioChainedInvocation = 'PerScenarioChainedInvocation'
    PerNodeInvocation =            'PerNodeInvocation'
    PerNodeChainedInvocation =     'PerNodeChainedInvocation'

class TransmitType(object):

    # Transmit stale variables
    stale       = 0b0000001
    # Transmit fixed variables
    fixed       = 0b0000010
    # Transmit derived variables
    derived     = 0b0000100
    # Transmit blended variables
    blended     = 0b0001000

    @classmethod
    def TransmitStale(cls, flag):
        return (flag & cls.stale) == cls.stale
    @classmethod
    def TransmitFixed(cls, flag):
        return (flag & cls.fixed) == cls.fixed
    @classmethod
    def TransmitDerived(cls, flag):
        return (flag & cls.derived) == cls.derived

    # Scenario tree variables
    nonleaf_stages     = 0b0010000
    all_stages         = 0b0110000

    @classmethod
    def TransmitNonLeafStages(cls, flag):
        return flag & cls.nonleaf_stages == cls.nonleaf_stages
    @classmethod
    def TransmitAllStages(cls, flag):
        return flag & cls.all_stages == cls.all_stages

def collect_full_results(ph, var_config):

    start_time = time.time()

    if ph._verbose:
        print("Collecting results from PH solver servers")

    subproblem_action_handle_map = {} # maps subproblem names to action handles
    action_handle_subproblem_map = {} # maps action handles to subproblem names

    ph._solver_manager.begin_bulk()

    for subproblem_to_solve in ph._scenario_tree.subproblems:

        new_action_handle =  ph._solver_manager.queue(
            action="collect_results",
            queue_name=ph._phpyro_job_worker_map[subproblem_to_solve.name],
            name=subproblem_to_solve.name,
            var_config=var_config)

        subproblem_action_handle_map[subproblem_to_solve.name] = new_action_handle
        action_handle_subproblem_map[new_action_handle] = subproblem_to_solve.name

    ph._solver_manager.end_bulk()

    if ph._verbose:
        print("Waiting for results extraction")

    num_results_so_far = 0

    while (num_results_so_far < len(ph._scenario_tree.subproblems)):

        action_handle = ph._solver_manager.wait_any()
        try:
            subproblem_name = action_handle_subproblem_map[action_handle]
        except KeyError:
            if action_handle in ph._queued_solve_action_handles:
                ph._queued_solve_action_handles.discard(action_handle)
                print("WARNING: Discarding uncollected solve action handle "
                      "with id=%d encountered during results collection"
                      % (action_handle.id))
                continue
            else:
                known_action_handles = \
                                       sorted((ah.id for ah in action_handle_subproblem_map))
                raise RuntimeError("PH client received an unknown action "
                                   "handle=%d from the dispatcher; known "
                                   "action handles are: %s"
                                   % (action_handle.id,
                                      str(known_action_handles)))

        subproblem_result = ph._solver_manager.get_results(action_handle)

        if ph._scenario_tree.is_bundle(subproblem_name):
            for scenario_name, scenario_results in iteritems(subproblem_result):
                scenario = ph._scenario_tree._scenario_map[scenario_name]
                scenario.set_solution(scenario_results)
        else:
            scenario = ph._scenario_tree._scenario_map[subproblem_name]
            scenario.set_solution(subproblem_result)

        if ph._verbose:
            print("Successfully loaded solution for subproblem="+subproblem_name)

        num_results_so_far += 1

    end_time = time.time()

    if ph._output_times:
        print("Results collection time=%.2f seconds" % (end_time - start_time))

#
# Extracts warm-starts from the solver servers.
#

def warmstart_scenario_instances(ph):

    start_time = time.time()

    if ph._verbose:
        print("Collecting warmstart from PH solver servers")

    scenario_action_handle_map = {} # maps scenario names to action handles
    action_handle_scenario_map = {} # maps action handles to scenario names

    ph._solver_manager.begin_bulk()

    if ph._scenario_tree.contains_bundles():

        for bundle in ph._scenario_tree._scenario_bundles:

            for scenario_name in bundle._scenario_names:

                new_action_handle =  ph._solver_manager.queue(
                    action="collect_warmstart",
                    queue_name=ph._phpyro_job_worker_map[bundle.name],
                    name=bundle.name,
                    scenario_name=scenario_name)

                scenario_action_handle_map[scenario_name] = new_action_handle
                action_handle_scenario_map[new_action_handle] = scenario_name

    else:

        for scenario in ph._scenario_tree._scenarios:

            new_action_handle = ph._solver_manager.queue(
                action="collect_warmstart",
                queue_name=ph._phpyro_job_worker_map[scenario.name],
                name=scenario.name,
                scenario_name=scenario.name)

            scenario_action_handle_map[scenario.name] = new_action_handle
            action_handle_scenario_map[new_action_handle] = scenario.name

    ph._solver_manager.end_bulk()

    if ph._verbose:
        print("Waiting for warmstart results")

    num_results_so_far = 0

    while (num_results_so_far < len(ph._scenario_tree._scenarios)):

        action_handle = ph._solver_manager.wait_any()
        try:
            scenario_name = action_handle_scenario_map[action_handle]
        except KeyError:
            if action_handle in ph._queued_solve_action_handles:
                ph._queued_solve_action_handles.discard(action_handle)
                print("WARNING: Discarding uncollected solve action handle "
                      "with id=%d encountered during scenario results collection"
                      % (action_handle.id))
                continue
            else:
                known_action_handles = \
                    sorted((ah.id for ah in action_handle_scenario_map))
                raise RuntimeError("PH client received an unknown action "
                                   "handle=%d from the dispatcher; known "
                                   "action handles are: %s"
                                   % (action_handle.id,
                                      str(known_action_handles)))

        scenario_results = ph._solver_manager.get_results(action_handle)
        scenario = ph._scenario_tree._scenario_map[scenario_name]
        var_sm_bySymbol = scenario._instance._PHInstanceSymbolMaps[Var].bySymbol
        for symbol, val in iteritems(scenario_results):
            var_sm_bySymbol[symbol].value = val

        if ph._verbose:
            print("Successfully loaded warmstart for scenario="+scenario_name)

        num_results_so_far += 1

    end_time = time.time()

    if ph._output_times:
        print("Warmstart collection time=%.2f seconds" % (end_time - start_time))

def transmit_weights(ph):

    start_time = time.time()

    if ph._verbose:
        print("Transmitting instance weights to PH solver servers")

    action_handles = []

    generate_responses = ph._handshake_with_phpyro

    ph._solver_manager.begin_bulk()

    if ph._scenario_tree.contains_bundles():

        for bundle in ph._scenario_tree._scenario_bundles:

            # map from scenario name to the corresponding weight map
            weights_to_transmit = {}

            for scenario in bundle._scenario_tree._scenarios:
                scenario_name = scenario.name

                # Skip the leaf nodes (scenario._w usually doesn't
                # store a value for variables on the leaf node)
                weights_to_transmit[scenario.name] = \
                    ph._scenario_tree._scenario_map[scenario_name]._w

            action_handles.append( ph._solver_manager.queue(
                action="load_weights",
                queue_name=ph._phpyro_job_worker_map[bundle.name],
                generateResponse=generate_responses,
                name=bundle.name,
                new_weights=weights_to_transmit) )

    else:

        for scenario in ph._scenario_tree._scenarios:

            # Skip the leaf nodes (scenario._w usually doesn't store a value
            # for variables on the leaf node)
            action_handles.append( ph._solver_manager.queue(
                action="load_weights",
                queue_name=ph._phpyro_job_worker_map[scenario.name],
                generateResponse=generate_responses,
                name=scenario.name,
                new_weights=scenario._w) )

    ph._solver_manager.end_bulk()

    if generate_responses:
        ph._solver_manager.wait_all(action_handles)

    end_time = time.time()

    if ph._output_times:
        print("Weight transmission time=%.2f seconds" % (end_time - start_time))

#
# a utility to transmit - across the PH solver manager - the
# the current xbar values for each non-leaf scenario tree node
#

def transmit_xbars(ph):

    start_time = time.time()

    if ph._verbose:
        print("Transmitting instance xbars to PH solver servers")

    action_handles = []

    generate_responses = ph._handshake_with_phpyro

    ph._solver_manager.begin_bulk()

    if ph._scenario_tree.contains_bundles():

        for bundle in ph._scenario_tree._scenario_bundles:

            xbars_to_transmit = {}
            # Skip the leaf nodes
            for stage in bundle._scenario_tree._stages[:-1]:
                for bundle_tree_node in stage._tree_nodes:
                    # The bundle scenariotree usually isn't populated
                    # with variable value data so we need to reference
                    # the original scenariotree node
                    primary_tree_node = \
                        ph._scenario_tree._tree_node_map[bundle_tree_node.name]
                    xbars_to_transmit[primary_tree_node.name] = \
                        primary_tree_node._xbars

            action_handles.append( ph._solver_manager.queue(
                action="load_xbars",
                queue_name=ph._phpyro_job_worker_map[bundle.name],
                generateResponse=generate_responses,
                name=bundle.name,
                new_xbars=xbars_to_transmit) )

    else:

        for scenario in ph._scenario_tree._scenarios:

            # Skip the leaf nodes
            xbars_to_transmit = {}
            xbars_to_transmit = dict((tree_node.name, tree_node._xbars) \
                                     for tree_node in scenario._node_list[:-1])

            action_handles.append( ph._solver_manager.queue(
                action="load_xbars",
                queue_name=ph._phpyro_job_worker_map[scenario.name],
                generateResponse=generate_responses,
                name=scenario.name,
                new_xbars=xbars_to_transmit) )

    ph._solver_manager.end_bulk()

    if generate_responses:
        ph._solver_manager.wait_all(action_handles)

    end_time = time.time()

    if ph._output_times:
        print("Xbar transmission time=%.2f seconds" % (end_time - start_time))


def _transmit_init(ph, worker_name, object_name):

    # both the dispatcher queue for initialization and the action name
    # are "initialize" - might be confusing, but hopefully not so
    # much.

    ah = ph._solver_manager.queue(
        action="initialize",
        queue_name=worker_name,
        name=worker_name,
        model_location=ph._scenario_tree._scenario_instance_factory._model_filename,
        data_location=ph._scenario_tree._scenario_instance_factory._scenario_tree_filename,
        objective_sense=ph._objective_sense_option,
        object_name=object_name,
        solver_type=ph._solver_type,
        solver_io=ph._solver_io,
        scenario_bundle_specification=ph._scenario_bundle_specification,
        create_random_bundles=ph._create_random_bundles,
        scenario_tree_random_seed=ph._scenario_tree_random_seed,
        default_rho=ph._rho,
        linearize_nonbinary_penalty_terms=\
        ph._linearize_nonbinary_penalty_terms,
        retain_quadratic_binary_terms=ph._retain_quadratic_binary_terms,
        breakpoint_strategy=ph._breakpoint_strategy,
        integer_tolerance=ph._integer_tolerance,
        output_solver_results=ph._output_solver_results,
        verbose=ph._verbose,
        compile_scenario_instances=ph._options.compile_scenario_instances)

    return ah

def release_phsolverservers(ph):

    if ph._verbose:
        print("Revoking PHPyroWorker job assignments")

    ph._solver_manager.begin_bulk()

    for job, worker in iteritems(ph._phpyro_job_worker_map):
        ph._solver_manager.queue(action="release",
                                 queue_name=ph._phpyro_job_worker_map[job],
                                 name=worker,
                                 object_name=job,
                                 generateResponse=False)

    ph._solver_manager.end_bulk()

    ph._phpyro_worker_jobs_map = {}
    ph._phpyro_job_worker_map = {}

def initialize_ph_solver_servers(ph):

    start_time = time.time()

    if ph._verbose:
        print("Transmitting initialization information to PH solver servers")

    if len(ph._solver_manager.server_pool) == 0:
        raise RuntimeError("No PHSolverServer processes have been acquired!")

    if ph._scenario_tree.contains_bundles():
        worker_jobs = [bundle.name for bundle in ph._scenario_tree._scenario_bundles]
    else:
        worker_jobs = [scenario.name for scenario in ph._scenario_tree._scenarios]

    action_handles = []
    ph._phpyro_worker_jobs_map = {}
    ph._phpyro_job_worker_map = {}
    for worker_name in itertools.cycle(ph._solver_manager.server_pool):
        if len(worker_jobs) == 0:
            break
        job_name = worker_jobs.pop()
        action_handles.append(_transmit_init(ph, worker_name, job_name))
        ph._phpyro_worker_jobs_map.setdefault(worker_name,[]).append(job_name)
        ph._phpyro_job_worker_map[job_name] = worker_name
    end_time = time.time()

    if ph._output_times:
        print("Initialization transmission time=%.2f seconds"
              % (end_time - start_time))

    #ph._solver_manager.wait_all(action_handles)
    return action_handles

#
# a utility to transmit to each PH solver server the current rho
# values for each problem instance.
#

def transmit_rhos(ph):

    start_time = time.time()

    if ph._verbose:
        print("Transmitting instance rhos to PH solver servers")

    action_handles = []

    generate_responses = ph._handshake_with_phpyro

    ph._solver_manager.begin_bulk()

    if ph._scenario_tree.contains_bundles():

        for bundle in ph._scenario_tree._scenario_bundles:

            # map from scenario name to the corresponding rho map
            rhos_to_transmit = {}

            for scenario in bundle._scenario_tree._scenarios:
                # Skip the leaf nodes (scenario._rho usually doesn't
                # store a value for variables on the leaf node)
                rhos_to_transmit[scenario.name] = \
                    ph._scenario_tree._scenario_map[scenario.name]._rho

            action_handles.append( ph._solver_manager.queue(
                action="load_rhos",
                queue_name=ph._phpyro_job_worker_map[bundle.name],
                name=bundle.name,
                generateResponse=generate_responses,
                new_rhos=rhos_to_transmit) )

    else:

        for scenario in ph._scenario_tree._scenarios:

            # Skip the leaf nodes (scenario._rho usually doesn't store
            # a value for variables on the leaf node)
            action_handles.append( ph._solver_manager.queue(
                action="load_rhos",
                queue_name=ph._phpyro_job_worker_map[scenario.name],
                name=scenario.name,
                generateResponse=generate_responses,
                new_rhos=scenario._rho) )

            ph._solver_manager.end_bulk()

    if generate_responses:
        ph._solver_manager.wait_all(action_handles)

    end_time = time.time()

    if ph._output_times:
        print("Rho transmission time=%.2f seconds" % (end_time - start_time))

#
# a utility to transmit - across the PH solver manager - the current
# scenario tree node statistics to each of my problem instances. done
# prior to each PH iteration k.
#

def transmit_tree_node_statistics(ph):

    start_time = time.time()

    if ph._verbose:
        print("Transmitting tree node statistics to PH solver servers")

    action_handles = []

    generate_responses = ph._handshake_with_phpyro

    ph._solver_manager.begin_bulk()

    if ph._scenario_tree.contains_bundles():

        for bundle in ph._scenario_tree._scenario_bundles:

            tree_node_minimums = {}
            tree_node_maximums = {}
            # iterate over the tree nodes in the bundle scenario tree - but
            # there aren't any statistics there - be careful!
            # TBD - we need to form these statistics! right now, they are
            #       beyond the bundle.
            # We ignore the leaf nodes
            for stage in bundle._scenario_tree._stages[:-1]:
                for bundle_tree_node in stage._tree_nodes:
                    primary_tree_node = \
                        ph._scenario_tree._tree_node_map[bundle_tree_node.name]
                    tree_node_minimums[primary_tree_node.name] = \
                        primary_tree_node._minimums
                    tree_node_maximums[primary_tree_node.name] = \
                        primary_tree_node._maximums

            action_handles.append( ph._solver_manager.queue(
                action="load_tree_node_stats",
                queue_name=ph._phpyro_job_worker_map[bundle.name],
                name=bundle.name,
                generateResponse=generate_responses,
                new_mins=tree_node_minimums,
                new_maxs=tree_node_maximums) )

    else:

        for scenario in ph._scenario_tree._scenarios:

            tree_node_minimums = {}
            tree_node_maximums = {}

            # Skip the leaf nodes
            for tree_node in scenario._node_list[:-1]:
                tree_node_minimums[tree_node.name] = tree_node._minimums
                tree_node_maximums[tree_node.name] = tree_node._maximums

            action_handles.append( ph._solver_manager.queue(
                action="load_tree_node_stats",
                queue_name=ph._phpyro_job_worker_map[scenario.name],
                name=scenario.name,
                generateResponse=generate_responses,
                new_mins=tree_node_minimums,
                new_maxs=tree_node_maximums) )
    ph._solver_manager.end_bulk()

    if generate_responses:
        ph._solver_manager.wait_all(action_handles)

    end_time = time.time()

    if ph._output_times:
        print("Tree node statistics transmission time="
              "%.2f seconds" % (end_time - start_time))

#
# a utility to activate - across the PH solver manager - weighted
# penalty objective terms.
#

def activate_ph_objective_weight_terms(ph):

    if ph._verbose:
        print("Transmitting request to activate PH objective "
              "weight terms to PH solver servers")

    action_handles = []

    generate_responses = ph._handshake_with_phpyro

    ph._solver_manager.begin_bulk()

    for subproblem in ph._scenario_tree.subproblems:
        action_handles.append( ph._solver_manager.queue(
            action="activate_ph_objective_weight_terms",
            queue_name=ph._phpyro_job_worker_map[subproblem.name],
            generateResponse=generate_responses,
            name=subproblem.name) )

    ph._solver_manager.end_bulk()

    if generate_responses:
        ph._solver_manager.wait_all(action_handles)

#
# a utility to deactivate - across the PH solver manager - weighted
# penalty objective terms.
#

def deactivate_ph_objective_weight_terms(ph):

    if ph._verbose:
        print("Transmitting request to deactivate PH objective "
              "weight terms to PH solver servers")

    action_handles = []

    generate_responses = ph._handshake_with_phpyro

    ph._solver_manager.begin_bulk()

    for subproblem in ph._scenario_tree.subproblems:
        action_handles.append( ph._solver_manager.queue(
            action="deactivate_ph_objective_weight_terms",
            queue_name=ph._phpyro_job_worker_map[subproblem.name],
            generateResponse=generate_responses,
            name=subproblem.name) )

        ph._solver_manager.end_bulk()

    if generate_responses:
        ph._solver_manager.wait_all(action_handles)


#
# a utility to activate - across the PH solver manager - proximal
# penalty objective terms.
#

def activate_ph_objective_proximal_terms(ph):

    if ph._verbose:
        print("Transmitting request to activate PH objective "
              "proximal terms to PH solver servers")

    action_handles = []

    generate_responses = ph._handshake_with_phpyro

    ph._solver_manager.begin_bulk()

    for subproblem in ph._scenario_tree.subproblems:
        action_handles.append( ph._solver_manager.queue(
            action="activate_ph_objective_proximal_terms",
            queue_name=ph._phpyro_job_worker_map[subproblem.name],
            generateResponse=generate_responses,
            name=subproblem.name) )

    ph._solver_manager.end_bulk()

    if generate_responses:
        ph._solver_manager.wait_all(action_handles)

#
# a utility to deactivate - across the PH solver manager - proximal
# penalty objective terms.
#

def deactivate_ph_objective_proximal_terms(ph):

    if ph._verbose:
        print("Transmitting request to deactivate PH objective "
              "proximal terms to PH solver servers")

    action_handles = []

    generate_responses = ph._handshake_with_phpyro

    ph._solver_manager.begin_bulk()

    for subproblem in ph._scenario_tree.subproblems:
        action_handles.append( ph._solver_manager.queue(
            action="deactivate_ph_objective_proximal_terms",
            queue_name=ph._phpyro_job_worker_map[subproblem.name],
            generateResponse=generate_responses,
            name=subproblem.name) )

    ph._solver_manager.end_bulk()

    if generate_responses:
        ph._solver_manager.wait_all(action_handles)


def transmit_fixed_variables(ph):

    start_time = time.time()

    if ph._verbose:
        print("Synchronizing fixed variable status with PH solver servers")

    action_handles = []

    generate_responses = ph._handshake_with_phpyro

    ph._solver_manager.begin_bulk()
    if ph._scenario_tree.contains_bundles():

        for bundle in ph._scenario_tree._scenario_bundles:

            transmit_variables = False
            for bundle_tree_node in bundle._scenario_tree._tree_nodes:
                    primary_tree_node = \
                        ph._scenario_tree._tree_node_map[bundle_tree_node.name]
                    if len(primary_tree_node._fix_queue):
                        transmit_variables = True
                        break

            if transmit_variables:
                # map from node name to the corresponding list of
                # fixed variables
                fixed_variables_to_transmit = {}

                # Just send the entire state of fixed variables
                # on each node (including leaf nodes)
                for bundle_tree_node in bundle._scenario_tree._tree_nodes:
                    primary_tree_node = \
                        ph._scenario_tree._tree_node_map[bundle_tree_node.name]
                    fixed_variables_to_transmit[primary_tree_node.name] = \
                        primary_tree_node._fix_queue

                action_handles.append( ph._solver_manager.queue(
                    action="update_fixed_variables",
                    queue_name=ph._phpyro_job_worker_map[bundle.name],
                    name=bundle.name,
                    generateResponse=generate_responses,
                    fixed_variables=fixed_variables_to_transmit) )
            else:
                if ph._verbose:
                    print("No synchronization was needed for bundle %s"
                          % (bundle.name))

    else:

        for scenario in ph._scenario_tree._scenarios:

            transmit_variables = False
            for tree_node in scenario._node_list:
                if len(tree_node._fix_queue):
                    transmit_variables = True
                    break

            if transmit_variables:

                fixed_variables_to_transmit = \
                    dict((tree_node.name, tree_node._fix_queue)
                         for tree_node in scenario._node_list)

                action_handles.append( ph._solver_manager.queue(
                    action="update_fixed_variables",
                    queue_name=ph._phpyro_job_worker_map[scenario.name],
                    name=scenario.name,
                    generateResponse=generate_responses,
                    fixed_variables=fixed_variables_to_transmit) )
            else:
                if ph._verbose:
                    print("No synchronization was needed for scenario %s"
                          % (scenario.name))
    ph._solver_manager.end_bulk()

    if generate_responses:
        ph._solver_manager.wait_all(action_handles)

    end_time = time.time()

    if ph._output_times:
        print("Fixed variable synchronization time="
              "%.2f seconds" % (end_time - start_time))

def transmit_external_function_invocation_to_worker(
        ph,
        worker_name,
        module_name,
        function_name,
        invocation_type=InvocationType.SingleInvocation,
        return_action_handle=False,
        function_args=None,
        function_kwds=None):

    if ph._verbose:
        print("Transmitting external function invocation request to PH "
              "solver server with name %s" % worker_name)

    generate_response = ph._handshake_with_phpyro or return_action_handle

    if ph._scenario_tree.contains_bundles():
        if worker_name not in ph._scenario_tree._scenario_bundle_map:
            raise ValueError("PH solver server with name %s does not exist"
                             % (worker_name))
    else:
        if worker_name not in ph._scenario_tree._scenario_map:
            raise ValueError("PH solver server with name %s does not exist"
                             % (worker_name))

    action_handle = ph._solver_manager.queue(action="invoke_external_function",
                                             queue_name=ph._phpyro_job_worker_map[worker_name],
                                             name=worker_name,
                                             invocation_type=invocation_type.value,
                                             generateResponse=generate_response,
                                             module_name=module_name,
                                             function_name=function_name,
                                             function_kwds=function_kwds,
                                             function_args=function_args)

    if generate_response and (not return_action_handle):
        ph._solver_manager.wait_all([action_handle])

    return action_handle if (return_action_handle) else None

def transmit_external_function_invocation(
        ph,
        module_name,
        function_name,
        invocation_type=InvocationType.SingleInvocation,
        return_action_handles=False,
        function_args=None,
        function_kwds=None):

    start_time = time.time()

    if ph._verbose:
        print("Transmitting external function invocation request "
              "to PH solver servers")

    action_handles = []

    generate_responses = ph._handshake_with_phpyro or return_action_handles

    ph._solver_manager.begin_bulk()
    if ph._scenario_tree.contains_bundles():

        for bundle in ph._scenario_tree._scenario_bundles:

            action_handles.append(
                ph._solver_manager.queue(
                    action="invoke_external_function",
                    queue_name=ph._phpyro_job_worker_map[bundle.name],
                    name=bundle.name,
                    invocation_type=invocation_type.value,
                    generateResponse=generate_responses,
                    module_name=module_name,
                    function_name=function_name,
                    function_kwds=function_kwds,
                    function_args=function_args))

    else:

        for scenario in ph._scenario_tree._scenarios:

            action_handles.append(
                ph._solver_manager.queue(
                    action="invoke_external_function",
                    queue_name=ph._phpyro_job_worker_map[scenario.name],
                    name=scenario.name,
                    invocation_type=invocation_type.value,
                    generateResponse=generate_responses,
                    module_name=module_name,
                    function_name=function_name,
                    function_kwds=function_kwds,
                    function_args=function_args))
    ph._solver_manager.end_bulk()

    if generate_responses and (not return_action_handles):
        ph._solver_manager.wait_all(action_handles)

    end_time = time.time()

    if ph._output_times:
        print("External function invocation request transmission "
              "time=%.2f seconds" % (end_time - start_time))

    return action_handles if (return_action_handles) else None

#
# a utility to define model-level import suffixes - across the PH
# solver manager, on all instances.
#

def define_import_suffix(ph, suffix_name):

    if ph._verbose:
        print("Transmitting request to define suffix=%s to PH "
              "solver servers" % (suffix_name))

    action_handles = []

    generate_responses = ph._handshake_with_phpyro

    ph._solver_manager.begin_bulk()

    for subproblem in ph._scenario_tree.subproblems:
        action_handles.append( ph._solver_manager.queue(
            action="define_import_suffix",
            queue_name=ph._phpyro_job_worker_map[subproblem.name],
            generateResponse=generate_responses,
            name=subproblem.name,
            suffix_name = suffix_name))

    ph._solver_manager.end_bulk()

    if generate_responses:
        ph._solver_manager.wait_all(action_handles)

#
# a utility to request that each PH solver server restore cached
# solutions to their scenario instances.
#

def restore_cached_scenario_solutions(ph, cache_id, release_cache):

    if ph._verbose:
        print("Transmitting request to restore cached scenario solutions "
              "to PH solver servers")

    action_handles = []

    generate_responses = ph._handshake_with_phpyro

    ph._solver_manager.begin_bulk()

    for subproblem in ph._scenario_tree.subproblems:
        action_handles.append( ph._solver_manager.queue(
            action="restore_cached_scenario_solutions",
            queue_name=ph._phpyro_job_worker_map[subproblem.name],
            cache_id=cache_id,
            release_cache=release_cache,
            generateResponse=generate_responses,
            name=subproblem.name) )

    ph._solver_manager.end_bulk()

    if generate_responses:
        ph._solver_manager.wait_all(action_handles)

#
# a utility to request that each PH solver server cache
# solutions to their scenario instances.
#

def cache_scenario_solutions(ph, cache_id):

    if ph._verbose:
        print("Transmitting request to cache scenario solutions "
              "to PH solver servers")

    action_handles = []

    generate_responses = ph._handshake_with_phpyro

    ph._solver_manager.begin_bulk()

    for subproblem in ph._scenario_tree.subproblems:
        action_handles.append( ph._solver_manager.queue(
            action="cache_scenario_solutions",
            queue_name=ph._phpyro_job_worker_map[subproblem.name],
            cache_id=cache_id,
            generateResponse=generate_responses,
            name=subproblem.name) )

    ph._solver_manager.end_bulk()

    if generate_responses:
        ph._solver_manager.wait_all(action_handles)

def gather_scenario_tree_data(ph, initialization_action_handles):

    start_time = time.time()

    if ph._verbose:
        print("Collecting scenario tree data from PH solver servers")

    # maps scenario names to action handles
    scenario_action_handle_map = {}
    # maps action handles to scenario names
    action_handle_scenario_map = {}

    # maps bundle names to action handles
    bundle_action_handle_map = {}
    # maps action handles to bundle names
    action_handle_bundle_map = {}

    need_node_data = \
        dict((tree_node.name,True) \
             for tree_node in ph._scenario_tree._tree_nodes)
    need_scenario_data = \
        dict((scenario.name,True) \
             for scenario in ph._scenario_tree._scenarios)

    ph._solver_manager.begin_bulk()
    if ph._scenario_tree.contains_bundles():

        for bundle in ph._scenario_tree._scenario_bundles:

            object_names = {}
            object_names['nodes'] = \
                [tree_node.name \
                 for scenario in bundle._scenario_tree._scenarios \
                 for tree_node in scenario._node_list \
                 if need_node_data[tree_node.name]]
            object_names['scenarios'] = \
                [scenario_name \
                 for scenario_name in bundle._scenario_names]

            new_action_handle =  ph._solver_manager.queue(
                action="collect_scenario_tree_data",
                queue_name=ph._phpyro_job_worker_map[bundle.name],
                name=bundle.name,
                tree_object_names=object_names)

            bundle_action_handle_map[bundle.name] = new_action_handle
            action_handle_bundle_map[new_action_handle] = bundle.name

            for node_name in object_names['nodes']:
                need_node_data[node_name] = False
            for scenario_name in object_names['scenarios']:
                need_scenario_data[scenario_name] = False

    else:

        for scenario in ph._scenario_tree._scenarios:

            object_names = {}
            object_names['nodes'] = \
                [tree_node.name for tree_node in scenario._node_list \
                 if need_node_data[tree_node.name]]
            object_names['scenarios'] = [scenario.name]

            new_action_handle = ph._solver_manager.queue(
                action="collect_scenario_tree_data",
                queue_name=ph._phpyro_job_worker_map[scenario.name],
                name=scenario.name,
                tree_object_names=object_names)

            scenario_action_handle_map[scenario.name] = new_action_handle
            action_handle_scenario_map[new_action_handle] = scenario.name

            for node_name in object_names['nodes']:
                need_node_data[node_name] = False
            for scenario_name in object_names['scenarios']:
                need_scenario_data[scenario_name] = False
    ph._solver_manager.end_bulk()

    assert all(not val for val in itervalues(need_node_data))
    assert all(not val for val in itervalues(need_scenario_data))

    have_node_data = \
        dict((tree_node.name,False) \
             for tree_node in ph._scenario_tree._tree_nodes)
    have_scenario_data = \
        dict((scenario.name,False) \
             for scenario in ph._scenario_tree._scenarios)

    if ph._verbose:
        print("Waiting for scenario tree data extraction")

    if ph._scenario_tree.contains_bundles():

        num_results_so_far = 0

        while (num_results_so_far < len(ph._scenario_tree._scenario_bundles)):

            action_handle = ph._solver_manager.wait_any()

            if action_handle in initialization_action_handles:
                initialization_action_handles.remove(action_handle)
                ph._solver_manager.get_results(action_handle)
                continue

            bundle_results = ph._solver_manager.get_results(action_handle)
            bundle_name = action_handle_bundle_map[action_handle]

            for tree_node_name, node_data in iteritems(bundle_results['nodes']):
                assert have_node_data[tree_node_name] == False
                have_node_data[tree_node_name] = True
                tree_node = ph._scenario_tree.get_node(tree_node_name)
                tree_node._variable_ids.update(node_data['_variable_ids'])
                tree_node._standard_variable_ids.update(node_data['_standard_variable_ids'])
                tree_node._variable_indices.update(node_data['_variable_indices'])
                tree_node._integer.update(node_data['_integer'])
                tree_node._binary.update(node_data['_binary'])
                tree_node._semicontinuous.update(node_data['_semicontinuous'])
                # these are implied
                tree_node._derived_variable_ids = \
                    set(tree_node._variable_ids)-tree_node._standard_variable_ids
                tree_node._name_index_to_id = \
                    dict((val,key) for key,val in iteritems(tree_node._variable_ids))

            for scenario_name, scenario_data in \
                  iteritems(bundle_results['scenarios']):
                assert have_scenario_data[scenario_name] == False
                have_scenario_data[scenario_name] = True
                scenario = ph._scenario_tree.get_scenario(scenario_name)
                scenario._objective_name = scenario_data['_objective_name']
                scenario._objective_sense = scenario_data['_objective_sense']
                # rhos may have been modified with rhosetter callback
                scenario._rho.update(scenario_data['_rho'])
                # initialize _w, _rho, and _x, keys after this loop

            if ph._verbose:
                print("Successfully loaded scenario tree data "
                      "for bundle="+bundle_name)

            num_results_so_far += 1

    else:

        num_results_so_far = 0

        while (num_results_so_far < len(ph._scenario_tree._scenarios)):

            action_handle = ph._solver_manager.wait_any()

            if action_handle in initialization_action_handles:
                initialization_action_handles.remove(action_handle)
                ph._solver_manager.get_results(action_handle)
                continue

            scenario_results = ph._solver_manager.get_results(action_handle)
            scenario_name = action_handle_scenario_map[action_handle]
            for tree_node_name, node_data in iteritems(scenario_results['nodes']):
                assert have_node_data[tree_node_name] == False
                have_node_data[tree_node_name] = True
                tree_node = ph._scenario_tree.get_node(tree_node_name)
                tree_node._variable_ids.update(node_data['_variable_ids'])
                tree_node._standard_variable_ids.update(node_data['_standard_variable_ids'])
                tree_node._variable_indices.update(node_data['_variable_indices'])
                tree_node._integer.update(node_data['_integer'])
                tree_node._binary.update(node_data['_binary'])
                tree_node._semicontinuous.update(node_data['_semicontinuous'])
                # these are implied
                tree_node._derived_variable_ids = \
                    set(tree_node._variable_ids)-tree_node._standard_variable_ids
                tree_node._name_index_to_id = \
                    dict((val,key) for key,val in iteritems(tree_node._variable_ids))

            for scenario_name, scenario_data in \
                  iteritems(scenario_results['scenarios']):
                assert have_scenario_data[scenario_name] == False
                have_scenario_data[scenario_name] = True
                scenario = ph._scenario_tree.get_scenario(scenario_name)
                scenario._objective_name = scenario_data['_objective_name']
                scenario._objective_sense = scenario_data['_objective_sense']
                # rhos may have been modified with rhosetter callback
                scenario._rho.update(scenario_data['_rho'])
                # initialize _w and _x keys after this loop

            if ph._verbose:
                print("Successfully loaded scenario tree data for "
                      "scenario="+scenario_name)

            num_results_so_far += 1

    assert all(have_node_data)
    assert all(have_scenario_data)

    for tree_node in ph._scenario_tree._tree_nodes:
        tree_node._minimums = dict.fromkeys(tree_node._variable_ids,0)
        tree_node._maximums = dict.fromkeys(tree_node._variable_ids,0)
        # this is the true variable average at the node (unmodified)
        tree_node._averages = dict.fromkeys(tree_node._variable_ids,0)
        # this is the xbar used in the PH objective.
        tree_node._xbars = dict.fromkeys(tree_node._standard_variable_ids,0.0)
        # this is the blend used in the PH objective
        tree_node._blend = dict.fromkeys(tree_node._standard_variable_ids,1)
        # For the dual ph algorithm
        tree_node._wbars = dict.fromkeys(tree_node._standard_variable_ids,None)
        for scenario in tree_node._scenarios:
            scenario._x[tree_node.name] = \
                dict.fromkeys(tree_node._variable_ids,None)
            if not tree_node.is_leaf_node():
                scenario._w[tree_node.name] = \
                    dict.fromkeys(tree_node._standard_variable_ids,0.0)

    if len(initialization_action_handles):
        if ph._verbose:
            print("Waiting on remaining PHSolverServer initializations")
        ph._solver_manager.wait_all(initialization_action_handles)
        while len(initialization_action_handles):
            initialization_action_handles.pop()

    end_time = time.time()

    if ph._output_times:
        print("Scenario tree data collection time=%.2f seconds"
              % (end_time - start_time))



import hashlib
import uuid

import six

pysp_namespace_hash = hashlib.md5('pysp'.encode())
def _compute_namespace(node_name):
    node_namespace_hash = pysp_namespace_hash.copy()
    node_namespace_hash.update(node_name.encode())
    return uuid.UUID(bytes=node_namespace_hash.digest())

#
# 32-bit family
#

_max_int32 = 2**31 - 1
_max_uint32 = 2**32 - 1

def _convert_range_zero_to_max_int32(x):
    assert x >= 0
    # use int for py2 to eliminate long
    # types when possible
    return int(x % (_max_int32+1))
def _convert_range_one_to_max_int32(x):
    v = _convert_range_zero_to_max_int32(x)
    # a collision happens for 0 and 1
    if v == 0:
        v += 1
    return v
def _convert_range_zero_to_max_uint32(x):
    assert x >= 0
    # use int for py2 to eliminate long
    # types when possible
    return int(x % (_max_uint32+1))
def _convert_range_one_to_max_uint32(x):
    v = _convert_range_zero_to_max_uint32(x)
    # a collision happens for 0 and 1
    if v == 0:
        v += 1
    return v

def scenario_tree_id_to_pint32(node_name, vid):
    """
    Maps a PySP scenario tree variable id paired with a node
    name to a positive integer that can be stored in a
    32-bit signed integer type.

    The mapping is unique with high probability for a set of
    scenario tree ids that is not too large.
    (see: birthday problem)
    """
    return _convert_range_one_to_max_int32(
        uuid.uuid5(_compute_namespace(node_name), vid).int)

def scenario_tree_id_to_nzint32(node_name, vid):
    """
    Maps a PySP scenario tree variable id paired with a node
    name to a nonnegative integer that can be stored in a
    32-bit signed integer type.

    The mapping is unique with high probability for a set of
    scenario tree ids that is not too large.
    (see: birthday problem)
    """
    return _convert_range_zero_to_max_int32(
        uuid.uuid5(_compute_namespace(node_name), vid).int)

def scenario_tree_id_to_puint32(node_name, vid):
    """
    Maps a PySP scenario tree variable id paired with a node
    name to a positive integer that can be stored in a
    32-bit signed integer type.

    The mapping is unique with high probability for a set of
    scenario tree ids that is not too large.
    (see: birthday problem)
    """
    return _convert_range_one_to_max_uint32(
        uuid.uuid5(_compute_namespace(node_name), vid).int)

def scenario_tree_id_to_nzuint32(node_name, vid):
    """
    Maps a PySP scenario tree variable id paired with a node
    name to a nonnegative integer that can be stored in a
    32-bit signed integer type.

    The mapping is unique with high probability for a set of
    scenario tree ids that is not too large.
    (see: birthday problem)
    """
    return _convert_range_zero_to_max_uint32(
        uuid.uuid5(_compute_namespace(node_name), vid).int)

#
# 64-bit family
#

_max_int64 = 2**63 - 1
_max_uint64 = 2**64 - 1

def _convert_range_zero_to_max_int64(x):
    assert x >= 0
    # use int for py2 to eliminate long
    # types when possible
    return int(x % (_max_int64+1))
def _convert_range_one_to_max_int64(x):
    v = _convert_range_zero_to_max_int64(x)
    # a collision happens for 0 and 1
    if v == 0:
        v += 1
    return v
def _convert_range_zero_to_max_uint64(x):
    assert x >= 0
    # use int for py2 to eliminate long
    # types when possible
    return int(x % (_max_uint64+1))
def _convert_range_one_to_max_uint64(x):
    v = _convert_range_zero_to_max_uint64(x)
    # a collision happens for 0 and 1
    if v == 0:
        v += 1
    return v

def scenario_tree_id_to_pint64(node_name, vid):
    """
    Maps a PySP scenario tree variable id paired with a node
    name to a positive integer that can be stored in a
    64-bit signed integer type.

    The mapping is unique with high probability for a set of
    scenario tree ids that is not too large.
    (see: birthday problem)
    """
    return _convert_range_one_to_max_int64(
        uuid.uuid5(_compute_namespace(node_name), vid).int)

def scenario_tree_id_to_nzint64(node_name, vid):
    """
    Maps a PySP scenario tree variable id paired with a node
    name to a nonnegative integer that can be stored in a
    64-bit signed integer type.

    The mapping is unique with high probability for a set of
    scenario tree ids that is not too large.
    (see: birthday problem)
    """
    return _convert_range_zero_to_max_int64(
        uuid.uuid5(_compute_namespace(node_name), vid).int)

def scenario_tree_id_to_puint64(node_name, vid):
    """
    Maps a PySP scenario tree variable id paired with a node
    name to a positive integer that can be stored in a
    64-bit signed integer type.

    The mapping is unique with high probability for a set of
    scenario tree ids that is not too large.
    (see: birthday problem)
    """
    return _convert_range_one_to_max_uint64(
        uuid.uuid5(_compute_namespace(node_name), vid).int)

def scenario_tree_id_to_nzuint64(node_name, vid):
    """
    Maps a PySP scenario tree variable id paired with a node
    name to a nonnegative integer that can be stored in a
    64-bit signed integer type.

    The mapping is unique with high probability for a set of
    scenario tree ids that is not too large.
    (see: birthday problem)
    """
    return _convert_range_zero_to_max_uint64(
        uuid.uuid5(_compute_namespace(node_name), vid).int)



#  _________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright (c) 2014 Sandia Corporation.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  This software is distributed under the BSD License.
#  _________________________________________________________________________

# TODO: have ddsip convert create symbols files for second stage
# TODO: parse second-stage solution and load into scenario tree workers
# TODO: objective, cost, stage_costs
# TODO: make output_scenario_tree_solution work

import io
import os
import sys
import time
import shutil
import logging
import traceback

import pyutilib.subprocess
import pyutilib.services

from pyomo.core import maximize
from pyomo.opt import (TerminationCondition,
                       SolverStatus,
                       SolutionStatus)
from pyomo.pysp.util.configured_object import PySPConfiguredObject
from pyomo.pysp.util.config import (PySPConfigValue,
                                    PySPConfigBlock,
                                    safe_register_common_option,
                                    safe_register_unique_option,
                                    safe_declare_common_option,
                                    safe_declare_unique_option,
                                    _domain_tuple_of_str_or_dict,
                                    _domain_must_be_str)
from pyomo.pysp.util.misc import (parse_command_line,
                                  launch_command)
from pyomo.pysp.scenariotree.manager import \
    (InvocationType,
     ScenarioTreeManagerFactory)
import pyomo.pysp.convert.ddsip
from pyomo.pysp.embeddedsp import EmbeddedSP
from pyomo.pysp.solvers.spsolver import (SPSolverResults,
                                         SPSolverFactory)
from pyomo.pysp.solvers.spsolvershellcommand import \
    SPSolverShellCommand

logger = logging.getLogger('pyomo.pysp')

thisfile = os.path.abspath(__file__)

_ddsip_group_label = "DDSIP Options"
_firststage_var_suffix = "__DDSIP_FIRSTSTAGE"

# maps the ddsip status code to tuples of the form
# (SolutionStatus, SolverStatus, TerminationCondition, message)
# feel free to modify if you have opinions one these
_ddsip_status_map = {}
# the process was terminated by the user
_ddsip_status_map[-1] = (SolutionStatus.unknown,
                         SolverStatus.aborted,
                         TerminationCondition.userInterrupt,
                         "Termination signal received.")
# the node limit has been reached
_ddsip_status_map[1] = (SolutionStatus.stoppedByLimit,
                        SolverStatus.ok,
                        TerminationCondition.maxEvaluations,
                        "Node limit reached (total number of nodes).")
# the gap (absolute or relative) has been reached
_ddsip_status_map[2] = (SolutionStatus.optimal,
                        SolverStatus.ok,
                        TerminationCondition.optimal,
                        "Gap reached.")
# the time limit has been reached
_ddsip_status_map[3] = (SolutionStatus.stoppedByLimit,
                        SolverStatus.ok,
                        TerminationCondition.maxTimeLimit,
                        "Time limit reached.")
# the maximal dispersion, i.e. the maximal difference of the
# first-stage components within all remaining front nodes,
# was less then the parameter NULLDISP (null dispersion)
_ddsip_status_map[4] = (SolutionStatus.optimal,
                        SolverStatus.ok,
                        TerminationCondition.minStepLength,
                        "Maximal dispersion equals zero.")
# the whole branching tree was backtracked.
_ddsip_status_map[5] = (SolutionStatus.optimal,
                        SolverStatus.ok,
                        TerminationCondition.minStepLength,
                        ("The whole branching tree was "
                         "backtracked. Probably due to MIP "
                         "gaps (see below) the specified "
                         "gap tolerance could not be reached."))
# no valid lower bound
_ddsip_status_map[6] = (SolutionStatus.unknown,
                        SolverStatus.warning,
                        TerminationCondition.invalidProblem,
                        "No valid lower bound.")
# problem infeasible
_ddsip_status_map[7] = (SolutionStatus.infeasible,
                        SolverStatus.warning,
                        TerminationCondition.infeasible,
                        "Problem infeasible.")
# problem unbounded
_ddsip_status_map[8] = (SolutionStatus.unbounded,
                        SolverStatus.warning,
                        TerminationCondition.unbounded,
                        "Problem unbounded.")

def _load_solution(manager,
                   scenario,
                   solution_filename,
                   info_filename,
                   firststage_symbols_filename,
                   secondstage_symbols_filename,
                   scenario_id):

    # parse the symbol map
    firststage_symbol_map = {}
    with open(firststage_symbols_filename) as f:
        for line in f:
            lp_symbol, scenario_tree_id = line.strip().split()
            firststage_symbol_map[lp_symbol] = scenario_tree_id
    secondstage_symbol_map = {}
    with open(secondstage_symbols_filename) as f:
        for line in f:
            lp_symbol, scenario_tree_id = line.strip().split()
            secondstage_symbol_map[lp_symbol] = scenario_tree_id

    x_firststage = scenario._x[scenario.node_list[0].name]
    assert scenario.node_list[0] is manager.scenario_tree.findRootNode()
    x_secondstage = scenario._x[scenario.node_list[1].name]
    with open(solution_filename, 'r') as f:
        line = f.readline()
        while line.strip() != "1. Best Solution":
            line = f.readline()
        line = f.readline()
        assert line.startswith("Variable name                Value")
        line = f.readline()
        assert line.startswith("-----------------------------------")
        line = f.readline().strip()
        while line != "":
            varlabel, varsol = line.split()
            x_firststage[firststage_symbol_map[varlabel]] = \
                float(varsol)
            line = f.readline().strip()
        while line != "4. Second-stage solutions":
            line = f.readline().strip()
        scenario_label = ("Scenario %d:" % scenario_id)
        line = f.readline().strip()
        while line != scenario_label:
            line = f.readline().strip()
        line = f.readline().strip()
        while (line != "") and (not line.startswith("Scenario ")):
            varlabel, varsol = line.split()
            if varlabel != "ONE_VAR_CONSTANT":
                x_secondstage[secondstage_symbol_map[varlabel]] = \
                    float(varsol)
            line = f.readline().strip()
    return x

class DDSIPSolver(SPSolverShellCommand, PySPConfiguredObject):

    @classmethod
    def _declare_options(cls, options=None):
        if options is None:
            options = PySPConfigBlock()
        safe_declare_unique_option(
            options,
            "firststage_suffix",
            PySPConfigValue(
                "__DDSIP_FIRSTSTAGE",
                domain=_domain_must_be_str,
                description=(
                    "The suffix used to identity first-stage "
                    "variables to DDSIP. Default is "
                    "'__DDSIP_FIRSTSTAGE'"
                ),
                doc=None,
                visibility=0),
            ap_group=_ddsip_group_label)
        safe_declare_unique_option(
            options,
            "config_file",
            PySPConfigValue(
                None,
                domain=_domain_must_be_str,
                description=(
                    "The name of a partial DDSIP configuration file "
                    "that contains option specifications unrelated to "
                    "the problem structure. If specified, the contents "
                    "of this file will be appended to the "
                    "configuration created by this solver interface. "
                    "Default is None."
                ),
                doc=None,
                visibility=0),
            ap_group=_ddsip_group_label)
        safe_declare_common_option(options,
                                   "verbose",
                                   ap_group=_ddsip_group_label)

        return options

    def __init__(self):
        super(DDSIPSolver, self).__init__(self.register_options())
        self.set_options_to_default()
        self._executable = "ddsip"

    def set_options_to_default(self):
        self._options = self.register_options()
        self._options._implicit_declaration = True

    @property
    def options(self):
        return self._options

    @property
    def name(self):
        return "ddsip"

    def _solve_impl(self,
                    sp,
                    output_solver_log=False,
                    **kwds):
        """
        Solve a stochastic program with the DDSIP solver.

        See the 'solve' method on the base class for
        additional keyword documentation.

        Args:
            sp: The stochastic program to solve.
            output_solver_log (bool): Stream the solver
                output during the solve.
            **kwds: Passed to the DDSIP file writer as I/O
              options (e.g., symbolic_solver_labels=True).

        Returns: A results object with information about the solution.
        """

        if len(sp.scenario_tree.stages) > 2:
            raise ValueError("DDSIP solver does not handle more "
                             "than 2 time-stages")

        #
        # Setup the DDSIP working directory
        #

        working_directory = self._create_tempdir("workdir")
        input_directory = os.path.join(working_directory,
                                       "ddsip_files")
        output_directory = os.path.join(input_directory,
                                        "sipout")

        logfile = self._files['logfile'] = \
                  os.path.join(working_directory,
                               "ddsip.log")

        os.makedirs(input_directory)
        assert os.path.exists(input_directory)
        assert not os.path.exists(output_directory)
        info_filename = os.path.join(output_directory,
                                     "sip.out")
        solution_filename = os.path.join(output_directory,
                                         "solution.out")

        #
        # Create the DDSIP input files
        #

        if self.get_option("verbose"):
            print("Writing solver files in directory: %s"
                  % (working_directory))

        input_files = pyomo.pysp.convert.ddsip.\
            convert_external(
                input_directory,
                self.options.firststage_suffix,
                sp,
                io_options=kwds)
        for key in input_files:
            self._add_tempfile(key, input_files[key])

        self._update_config(input_files["config"])

        #
        # Launch DDSIP
        #

        _cmd_string = self.executable+" < "+input_files["script"]
        if self.get_option("verbose"):
            print("Launching DDSIP solver with command: %s"
                  % (_cmd_string))
        ddsipstdin = None
        with open(input_files["script"]) as f:
            ddsipstdin = f.read()
        assert ddsipstdin is not None

        start = time.time()
        rc, log = pyutilib.subprocess.run(
            self.executable,
            cwd=input_directory,
            stdin=ddsipstdin,
            outfile=logfile,
            tee=output_solver_log)
        stop = time.time()

        if rc:
            if not self.available():
                raise ValueError(
                    "Solver executable does not exist: '%s'. "
                    "(note that the default executable generated "
                    "by the DDSIP build system does not have this name)"
                    % (self.executable))
            else:
                raise RuntimeError(
                    "A nonzero return code (%s) was encountered after "
                    "launching the command: %s. Check the solver log file "
                    "for more information: %s"
                    % (rc, _cmd_string, logfile))

        #
        # Parse the DDSIP solution
        #

        if self.get_option("verbose"):
            print("Reading DDSIP solution from file: %s"
                  % (solution_filename))
        assert os.path.exists(output_directory)

        async_xhat = None
        async_responses = []

        # TODO: The solution symbols are not being translated
        #       back to scenario tree ids. Fix this and also have
        #       the first and second stage solutions be loaded into
        #       the Pyomo models
        #for scenario_id, scenario in enumerate(sp.scenario_tree.scenarios, 1):
        #    async_responses.append(sp.invoke_function(
        #        "_load_solution",
        #        thisfile,
        #        invocation_type=InvocationType.OnScenario(scenario.name),
        #        function_args=(solution_filename, scenario_id),
        #        async_call=True))

        xhat, results = self._read_solution(sp,
                                            input_files["symbols"],
                                            info_filename,
                                            solution_filename)

        results.xhat = None
        if xhat is not None:
            results.xhat = {sp.scenario_tree.findRootNode().name: xhat}

        for res in async_responses:
            res.complete()

        return results

    def _update_config(self, config_filename):
        """ Writes a DDSIP config file """

        # find the byte position where
        # we start appending to the config file
        # (just before END)
        append_pos = 0
        with io.open(config_filename,
                     mode='rb',
                     buffering=0) as f:
            f.seek(0)
            append_pos = f.tell()
            for line in f:
                if line.strip().decode() == "END":
                    break
                append_pos = f.tell()
        assert append_pos > 0

        config_lines = {}
        config_lines[None] = []
        config_lines['CPLEX'] = []
        config_lines['CPLEXEEV'] = []
        config_lines['CPLEXLB'] = []
        config_lines['CPLEX2LB'] = []
        config_lines['CPLEXUB'] = []
        config_lines['CPLEX2UB'] = []
        config_lines['CPLEXDUAL'] = []
        config_lines['CPLEX2DUAL'] = []
        has_cplex_opts = False

        # parse the user specified config file
        if self.options.config_file is not None:
            with open(self.options.config_file) as fconfig:
                section = config_lines[None]
                for line in fconfig:
                    stripped = line.strip()
                    if (stripped == "BEGIN") and \
                       (stripped == "END"):
                        continue
                    if "CPLEXBEGIN" in stripped:
                        section = config_lines['CPLEX']
                        has_cplex_opts = True
                        continue
                    elif "CPLEXEEV" in stripped:
                        section = config_lines['CPLEXEEV']
                        has_cplex_opts = True
                        continue
                    elif "CPLEXLB" in stripped:
                        section = config_lines['CPLEXLB']
                        has_cplex_opts = True
                        continue
                    elif "CPLEX2LB" in stripped:
                        section = config_lines['CPLEX2LB']
                        has_cplex_opts = True
                        continue
                    elif "CPLEXUB" in stripped:
                        section = config_lines['CPLEXUB']
                        has_cplex_opts = True
                        continue
                    elif "CPLEX2UB" in stripped:
                        section = config_lines['CPLEX2UB']
                        has_cplex_opts = True
                        continue
                    elif "CPLEXDUAL" in stripped:
                        section = config_lines['CPLEXDUAL']
                        has_cplex_opts = True
                        continue
                    elif "CPLEX2DUAL" in stripped:
                        section = config_lines['CPLEX2DUAL']
                        has_cplex_opts = True
                        continue
                    elif "CPLEXEND" in stripped:
                        section = config_lines[None]
                        continue
                    section.append(line)

        for key in self.options:
            if (key != "firststage_suffix") and \
               (key != "config_file") and \
               (key != "verbose"):
                if key.startswith("CPLEX_"):
                    section = config_lines["CPLEX"]
                    prefix = "CPLEX_"
                    has_cplex_opts = True
                elif key.startswith("CPLEXEEV_"):
                    section = config_lines["CPLEXEEV"]
                    prefix = "CPLEXEEV_"
                    has_cplex_opts = True
                elif key.startswith("CPLEXLB_"):
                    section = config_lines["CPLEXLB"]
                    prefix = "CPLEXLB_"
                    has_cplex_opts = True
                elif key.startswith("CPLEX2LB_"):
                    section = config_lines["CPLEX2LB"]
                    prefix = "CPLEX2LB_"
                    has_cplex_opts = True
                elif key.startswith("CPLEXUB_"):
                    section = config_lines["CPLEXUB"]
                    prefix = "CPLEXUB_"
                    has_cplex_opts = True
                elif key.startswith("CPLEX2UB_"):
                    section = config_lines["CPLEX2UB"]
                    prefix = "CPLEX2UB_"
                    has_cplex_opts = True
                elif key.startswith("CPLEXDUAL_"):
                    section = config_lines["CPLEXDUAL"]
                    prefix = "CPLEXDUAL_"
                    has_cplex_opts = True
                elif key.startswith("CPLEX2DUAL_"):
                    section = config_lines["CPLEX2DUAL"]
                    prefix = "CPLEX2DUAL_"
                    has_cplex_opts = True
                else:
                    section = config_lines[None]
                    prefix = ""
                val = self.options[key]
                line = "%s %s\n" % (key.replace(prefix,'',1), val)
                section.append(line)

        with open(config_filename, "r+") as f:
            f.seek(append_pos)
            for line in config_lines[None]:
                f.write(line)
            if has_cplex_opts:
                f.write("\nCPLEXBEGIN\n")
                for line in config_lines["CPLEX"]:
                    f.write(line)
                for section_key in ('CPLEXEEV',
                                    'CPLEXLB',
                                    'CPLEX2LB',
                                    'CPLEXUB',
                                    'CPLEX2UB',
                                    'CPLEXDUAL',
                                    'CPLEX2DUAL'):
                    section = config_lines[section_key]
                    if len(section) > 0:
                        f.write(section_key+"\n")
                        for line in section:
                            f.write(line)
                f.write("CPLEXEND\n")

            f.write("\n\nEND\n")

    def _read_solution(self,
                       sp,
                       symbols_filename,
                       info_filename,
                       solution_filename):
        """Parses a DDSIP solution file."""

        # parse the symbol map
        symbol_map = {}
        with open(symbols_filename) as f:
            for line in f:
                lp_symbol, scenario_tree_id = line.strip().split()
                symbol_map[lp_symbol] = scenario_tree_id

        #
        # Xhat
        #
        try:
            xhat = {}
            with open(solution_filename, 'r') as f:
                line = f.readline()
                while line.strip() != "1. Best Solution":
                    line = f.readline()
                line = f.readline()
                assert line.startswith("Variable name                Value")
                line = f.readline()
                assert line.startswith("-----------------------------------")
                line = f.readline().strip()
                while line != "":
                    line = line.split()
                    varlabel, varsol = line
                    xhat[symbol_map[varlabel]] = float(varsol)
                    line = f.readline().strip()
        except (IOError, OSError):
            logger.warn(
                "Exception encountered while parsing ddsip "
                "solution file '%s':\n%s'"
                % (solution_filename, traceback.format_exc()))
            xhat = None

        #
        # Objective, bound, status, etc.
        #
        results = SPSolverResults()
        results.solver.status_code = None
        results.status = None
        results.solver.status = None
        results.solver.termination_condition = None
        results.solver.message = None
        results.solver.time = None
        try:
            with open(info_filename, 'r') as f:
                line = f.readline()
                while True:
                    if line.startswith("Total CPU time:"):
                        break
                    line = f.readline()
                    if line == '':
                        # Unexpected file format or the solve failed
                        logger.warn(
                            "Unexpected ddsip info file format. No "
                            "status information will be returned")
                        return xhat, results
                line = f.readline().strip()
                while (line == "") or \
                      (line == "----------------------------------------------------------------------------------------"):
                    line = f.readline().strip()
                if line.startswith("EEV"):
                    results.eev = float(line.split()[1])
                    line = f.readline().strip()
                if line.startswith("VSS"):
                    results.vss = float(line.split()[1])
                    line = f.readline().strip()
                assert line.startswith("EVPI")
                line = line.split()
                results.evpi = float(line[1])
                line = f.readline().strip()
                assert line == ""
                line = f.readline().strip()
                assert line == "----------------------------------------------------------------------------------------"
                line = f.readline().strip()
                line = line.split()
                assert len(line) == 4
                assert line[0] == "Status"
                results.solver.status_code = int(line[1])
                assert line[2] == "Time"
                results.solver.time = float(line[3])
                (results.status,
                 results.solver.status,
                 results.solver.termination_condition,
                 results.solver.message) = \
                    _ddsip_status_map.get(results.solver.status_code,
                                          (SolutionStatus.unknown,
                                           SolverStatus.unknown,
                                           TerminationCondition.unknown,
                                           None))

                line = f.readline().strip()
                line = line.split()
                assert len(line) == 6
                assert line[0] == "Upper"
                assert line[3] == "Tree"
                results.tree_depth = int(line[5])
                line = f.readline().strip()
                line = line.split()
                assert len(line) == 2
                assert line[0] == "Nodes"
                results.nodes = int(line[1])
                line = f.readline().strip()
                while line != "----------------------------------------------------------------------------------------":
                    line = f.readline().strip()
                line = f.readline().strip()
                assert line.startswith("Best Value")
                results.objective = float(line.split()[2])
                # NOTE: I think DDSIP refers to the "bound" as
                #       "Lower Bound", even when the objective
                #       is being maximized.
                line = f.readline().strip()
                assert line.startswith("Lower Bound")
                results.bound = float(line.split()[2])
        except (IOError, OSError):
            logger.warn(
                "Exception encountered while parsing ddsip "
                "info file '%s':\n%s'"
                % (info_filename, traceback.format_exc()))

        return xhat, results

def runddsip_register_options(options=None):
    if options is None:
        options = PySPConfigBlock()
    DDSIPSolver.register_options(options)
    ScenarioTreeManagerFactory.register_options(options)
    safe_register_common_option(options,
                               "verbose")
    safe_register_common_option(options,
                               "disable_gc")
    safe_register_common_option(options,
                               "profile")
    safe_register_common_option(options,
                               "traceback")
    safe_register_common_option(options,
                                "output_scenario_tree_solution")
    safe_register_common_option(options,
                                "keep_solver_files")
    safe_register_common_option(options,
                                "output_solver_log")
    safe_register_common_option(options,
                                "symbolic_solver_labels")
    # used to populate the implicit DDSIP options
    safe_register_unique_option(
        options,
        "solver_options",
        PySPConfigValue(
            (),
            domain=_domain_tuple_of_str_or_dict,
            description=(
                "Unregistered solver options that will be passed "
                "to DDSIP via the config file (e.g., NODELIM=4, "
                "CPLEX_1067=1). This option can be used multiple "
                "times from the command line to specify more "
                "than one DDSIP option."
            ),
            doc=None,
            visibility=0),
        ap_kwds={'action': 'append'},
        ap_group=_ddsip_group_label)

    return options

def runddsip(options):
    """
    Construct a senario tree manager and solve it
    with the DDSIP solver.
    """
    start_time = time.time()
    with ScenarioTreeManagerFactory(options) as sp:
        sp.initialize()
        print("")
        print("Running DDSIP solver for stochastic "
              "programming problems")
        ddsip = DDSIPSolver()
        # add the implicit ddsip options
        solver_options = options.solver_options
        if len(solver_options) > 0:
            if type(solver_options) is tuple:
                for name_val in solver_options:
                    assert "=" in name_val
                    name, val = name_val.split("=")
                    ddsip.options[name.strip()] = val.strip()
            else:
                for key, val in solver_options.items():
                    ddsip.options[key] = val
        ddsip_options = ddsip.extract_user_options_to_dict(options,
                                                           sparse=True)
        results = ddsip.solve(
            sp,
            options=ddsip_options,
            output_solver_log=options.output_solver_log,
            keep_solver_files=options.keep_solver_files,
            symbolic_solver_labels=options.symbolic_solver_labels)
        xhat = results.xhat
        del results.xhat
        print("")
        print(results)

        if options.output_scenario_tree_solution:
            print("")
            sp.scenario_tree.snapshotSolutionFromScenarios()
            sp.scenario_tree.pprintSolution()
            sp.scenario_tree.pprintCosts()

    print("")
    print("Total execution time=%.2f seconds"
          % (time.time() - start_time))
    return 0

#
# the main driver routine
#

def main(args=None):
    #
    # Top-level command that executes everything
    #

    #
    # Import plugins
    #
    import pyomo.environ

    #
    # Parse command-line options.
    #
    try:
        options = parse_command_line(
            args,
            runddsip_register_options,
            prog='runddsip',
            description=(
"""Optimize a stochastic program using the DDSIP solver."""
            ))

    except SystemExit as _exc:
        # the parser throws a system exit if "-h" is specified
        # - catch it to exit gracefully.
        return _exc.code

    return launch_command(runddsip,
                          options,
                          error_label="runddsip: ",
                          disable_gc=options.disable_gc,
                          profile_count=options.profile,
                          traceback=options.traceback)

SPSolverFactory.register_solver("ddsip", DDSIPSolver)

if __name__ == "__main__":
    sys.exit(main())



import math
import warnings
import logging
from pyomo.common.errors import DeveloperError, InfeasibleConstraintException, PyomoException

logger = logging.getLogger(__name__)
inf = float('inf')


class IntervalException(PyomoException):
    pass


def add(xl, xu, yl, yu):
    return xl + yl, xu + yu


def sub(xl, xu, yl, yu):
    return xl - yu, xu - yl


def mul(xl, xu, yl, yu):
    lb = min(xl*yl, xl*yu, xu*yl, xu*yu)
    ub = max(xl*yl, xl*yu, xu*yl, xu*yu)
    if math.isnan(lb):
        lb = -inf
    if math.isnan(ub):
        ub = inf
    return lb, ub


def inv(xl, xu, feasibility_tol):
    if xl <= feasibility_tol and xu >= -feasibility_tol:
        # if the denominator (x) can include 0, then 1/x is unbounded.
        lb = -inf
        ub = inf
    else:
        ub = 1.0 / xl
        lb = 1.0 / xu
    return lb, ub


def div(xl, xu, yl, yu, feasibility_tol):
    if yl <= feasibility_tol and yu >= -feasibility_tol:
        # if the denominator (y) can include 0, then x/y is unbounded.
        lb = -inf
        ub = inf
    else:
        lb, ub = mul(xl, xu, *inv(yl, yu, feasibility_tol))
    return lb, ub


def power(xl, xu, yl, yu):
    """
    Compute bounds on x**y.
    """
    if xl > 0:
        """
        If x is always positive, things are simple. We only need to worry about the sign of y.
        """
        if yl < 0 and yu > 0:
            lb = min(xu ** yl, xl ** yu)
            ub = max(xl ** yl, xu ** yu)
        elif yl >= 0:
            lb = min(xl**yl, xl**yu)
            ub = max(xu**yl, xu**yu)
        elif yu <= 0:
            lb = min(xu**yl, xu**yu)
            ub = max(xl**yl, xl**yu)
        else:
            raise DeveloperError()
    elif xl == 0:
        if yl >= 0:
            lb = min(xl ** yl, xl ** yu)
            ub = max(xu ** yl, xu ** yu)
        else:
            lb = -inf
            ub = inf
    elif yl == yu and yl == round(yl):
        # the exponent is an integer, so x can be negative
        """
        The logic here depends on several things:
        1) The sign of x
        2) The sign of y
        3) Whether y is even or odd.
        
        There are also special cases to avoid math domain errors.
        """
        y = yl
        if xu <= 0:
            if y < 0:
                if y % 2 == 0:
                    lb = xl ** y
                    if xu == 0:
                        ub = inf
                    else:
                        ub = xu ** y
                else:
                    if xu == 0:
                        lb = -inf
                        ub = inf
                    else:
                        lb = xu ** y
                        ub = xl ** y
            else:
                if y % 2 == 0:
                    lb = xu ** y
                    ub = xl ** y
                else:
                    lb = xl ** y
                    ub = xu ** y
        else:
            if y < 0:
                if y % 2 == 0:
                    lb = min(xl ** y, xu ** y)
                    ub = inf
                else:
                    lb = - inf
                    ub = inf
            else:
                if y % 2 == 0:
                    lb = 0
                    ub = max(xl ** y, xu ** y)
                else:
                    lb = xl ** y
                    ub = xu ** y
    elif yl == yu:
        # the exponent has to be fractional, so x must be positive
        if xu < 0:
            msg = 'Cannot raise a negative number to the power of {0}.\n'.format(yl)
            msg += 'The upper bound of a variable raised to the power of {0} is {1}'.format(yl, xu)
            raise InfeasibleConstraintException(msg)
        xl = 0
        lb, ub = power(xl, xu, yl, yu)
    else:
        lb = -inf
        ub = inf

    return lb, ub


def _inverse_power1(zl, zu, yl, yu, orig_xl, orig_xu, feasibility_tol):
    """
    z = x**y => compute bounds on x.

    First, start by computing bounds on x with

        x = exp(ln(z) / y)

    However, if y is an integer, then x can be negative, so there are several special cases. See the docs below.
    """
    xl, xu = log(zl, zu)
    xl, xu = div(xl, xu, yl, yu, feasibility_tol)
    xl, xu = exp(xl, xu)

    # if y is an integer, then x can be negative
    if yl == yu and yl == round(yl):  # y is a fixed integer
        y = yl
        if y == 0:
            # Anything to the power of 0 is 1, so if y is 0, then x can be anything
            # (assuming zl <= 1 <= zu, which is enforced when traversing the tree in the other direction)
            xl = -inf
            xu = inf
        elif y % 2 == 0:
            """
            if y is even, then there are two primary cases (note that it is much easier to walk through these
            while looking at plots):
            case 1: y is positive
                x**y is convex, positive, and symmetric. The bounds on x depend on the lower bound of z. If zl <= 0, 
                then xl should simply be -xu. However, if zl > 0, then we may be able to say something better. For 
                example, if the original lower bound on x is positive, then we can keep xl computed from 
                x = exp(ln(z) / y). Furthermore, if the original lower bound on x is larger than -xl computed from 
                x = exp(ln(z) / y), then we can still keep the xl computed from x = exp(ln(z) / y). Similar logic
                applies to the upper bound of x.
            case 2: y is negative
                The ideas are similar to case 1.
            """
            if zu + feasibility_tol < 0:
                raise InfeasibleConstraintException('Infeasible. Anything to the power of {0} must be positive.'.format(y))
            if y > 0:
                if zu <= 0:
                    _xl = 0
                    _xu = 0
                elif zl <= 0:
                    _xl = -xu
                    _xu = xu
                else:
                    if orig_xl <= -xl + feasibility_tol:
                        _xl = -xu
                    else:
                        _xl = xl
                    if orig_xu < xl - feasibility_tol:
                        _xu = -xl
                    else:
                        _xu = xu
                xl = _xl
                xu = _xu
            else:
                if zu == 0:
                    raise InfeasibleConstraintException('Infeasible. Anything to the power of {0} must be positive.'.format(y))
                elif zl <= 0:
                    _xl = -inf
                    _xu = inf
                else:
                    if orig_xl <= -xl + feasibility_tol:
                        _xl = -xu
                    else:
                        _xl = xl
                    if orig_xu < xl - feasibility_tol:
                        _xu = -xl
                    else:
                        _xu = xu
                xl = _xl
                xu = _xu
        else:  # y % 2 == 1
            """
            y is odd. 
            Case 1: y is positive
                x**y is monotonically increasing. If y is positive, then we can can compute the bounds on x using
                x = z**(1/y) and the signs on xl and xu depend on the signs of zl and zu.
            Case 2: y is negative
                Again, this is easier to visualize with a plot. x**y approaches zero when x approaches -inf or inf. 
                Thus, if zl < 0 < zu, then no bounds can be inferred for x. If z is positive (zl >=0 ) then we can
                use the bounds computed from x = exp(ln(z) / y). If z is negative (zu <= 0), then we live in the
                bottom left quadrant, xl depends on zu, and xu depends on zl.
            """
            if y > 0:
                xl = abs(zl)**(1.0/y)
                xl = math.copysign(xl, zl)
                xu = abs(zu)**(1.0/y)
                xu = math.copysign(xu, zu)
            else:
                if zl >= 0:
                    pass
                elif zu <= 0:
                    if zu == 0:
                        xl = -inf
                    else:
                        xl = -abs(zu)**(1.0/y)
                    if zl == 0:
                        xu = -inf
                    else:
                        xu = -abs(zl)**(1.0/y)
                else:
                    xl = -inf
                    xu = inf

    return xl, xu


def _inverse_power2(zl, zu, xl, xu, feasiblity_tol):
    """
    z = x**y => compute bounds on y
    y = ln(z) / ln(x)

    This function assumes the exponent can be fractional, so x must be positive. This method should not be called
    if the exponent is an integer.
    """
    if xu <= 0:
        raise IntervalException('Cannot raise a negative variable to a fractional power.')
    if (xl > 0 and zu <= 0) or (xl >= 0 and zu < 0):
        raise InfeasibleConstraintException('A positive variable raised to the power of anything must be positive.')
    lba, uba = log(zl, zu)
    lbb, ubb = log(xl, xu)
    yl, yu = div(lba, uba, lbb, ubb, feasiblity_tol)
    return yl, yu


def exp(xl, xu):
    try:
        lb = math.exp(xl)
    except OverflowError:
        lb = math.inf
    try:
        ub = math.exp(xu)
    except OverflowError:
        ub = math.inf
    return lb, ub


def log(xl, xu):
    if xl > 0:
        lb = math.log(xl)
    else:
        lb = -inf
    if xu > 0:
        ub = math.log(xu)
    else:
        ub = -inf
    return lb, ub


def log10(xl, xu):
    if xl > 0:
        lb = math.log10(xl)
    else:
        lb = -inf
    if xu > 0:
        ub = math.log10(xu)
    else:
        ub = -inf
    return lb, ub


def sin(xl, xu):
    """

    Parameters
    ----------
    xl: float
    xu: float

    Returns
    -------
    lb: float
    ub: float
    """

    # if there is a minimum between xl and xu, then the lower bound is -1. Minimums occur at 2*pi*n - pi/2
    # find the minimum value of i such that 2*pi*i - pi/2 >= xl. Then round i up. If 2*pi*i - pi/2 is still less
    # than or equal to xu, then there is a minimum between xl and xu. Thus the lb is -1. Otherwise, the minimum
    # occurs at either xl or xu
    if xl <= -inf or xu >= inf:
        return -1, 1
    pi = math.pi
    i = (xl + pi / 2) / (2 * pi)
    i = math.ceil(i)
    x_at_min = 2 * pi * i - pi / 2
    if x_at_min <= xu:
        lb = -1
    else:
        lb = min(math.sin(xl), math.sin(xu))

    # if there is a maximum between xl and xu, then the upper bound is 1. Maximums occur at 2*pi*n + pi/2
    i = (xu - pi / 2) / (2 * pi)
    i = math.floor(i)
    x_at_max = 2 * pi * i + pi / 2
    if x_at_max >= xl:
        ub = 1
    else:
        ub = max(math.sin(xl), math.sin(xu))

    return lb, ub


def cos(xl, xu):
    """

    Parameters
    ----------
    xl: float
    xu: float

    Returns
    -------
    lb: float
    ub: float
    """

    # if there is a minimum between xl and xu, then the lower bound is -1. Minimums occur at 2*pi*n - pi
    # find the minimum value of i such that 2*pi*i - pi >= xl. Then round i up. If 2*pi*i - pi/2 is still less
    # than or equal to xu, then there is a minimum between xl and xu. Thus the lb is -1. Otherwise, the minimum
    # occurs at either xl or xu
    if xl <= -inf or xu >= inf:
        return -1, 1
    pi = math.pi
    i = (xl + pi) / (2 * pi)
    i = math.ceil(i)
    x_at_min = 2 * pi * i - pi
    if x_at_min <= xu:
        lb = -1
    else:
        lb = min(math.cos(xl), math.cos(xu))

    # if there is a maximum between xl and xu, then the upper bound is 1. Maximums occur at 2*pi*n
    i = (xu) / (2 * pi)
    i = math.floor(i)
    x_at_max = 2 * pi * i
    if x_at_max >= xl:
        ub = 1
    else:
        ub = max(math.cos(xl), math.cos(xu))

    return lb, ub


def tan(xl, xu):
    """

    Parameters
    ----------
    xl: float
    xu: float

    Returns
    -------
    lb: float
    ub: float
    """

    # tan goes to -inf and inf at every pi*i + pi/2 (integer i). If one of these values is between xl and xu, then
    # the lb is -inf and the ub is inf. Otherwise the minimum occurs at xl and the maximum occurs at xu.
    # find the minimum value of i such that pi*i + pi/2 >= xl. Then round i up. If pi*i + pi/2 is still less
    # than or equal to xu, then there is an undefined point between xl and xu.
    if xl <= -inf or xu >= inf:
        return -inf, inf
    pi = math.pi
    i = (xl - pi / 2) / (pi)
    i = math.ceil(i)
    x_at_undefined = pi * i + pi / 2
    if x_at_undefined <= xu:
        lb = -inf
        ub = inf
    else:
        lb = math.tan(xl)
        ub = math.tan(xu)

    return lb, ub


def asin(xl, xu, yl, yu, feasibility_tol):
    """
    y = asin(x); propagate bounds from x to y
    x = sin(y)

    Parameters
    ----------
    xl: float
    xu: float
    yl: float
    yu: float

    Returns
    -------
    lb: float
    ub: float
    """
    if xl < -1:
        xl = -1
    if xu > 1:
        xu = 1

    pi = math.pi

    if yl <= -inf:
        lb = yl
    elif xl <= math.sin(yl) <= xu:
        # if sin(yl) >= xl then yl satisfies the bounds on x, and the lower bound of y cannot be improved
        lb = yl
    elif math.sin(yl) < xl:
        """
        we can only push yl up from its current value to the next lowest value such that xl = sin(y). In other words,
        we need to

        min y
        s.t.
            xl = sin(y)
            y >= yl

        globally. 
        """
        # first find the next minimum of x = sin(y). Minimums occur at y = 2*pi*n - pi/2 for integer n.
        i = (yl + pi / 2) / (2 * pi)
        i1 = math.floor(i)
        i2 = math.ceil(i)
        i1 = 2 * pi * i1 - pi / 2
        i2 = 2 * pi * i2 - pi / 2
        # now find the next value of y such that xl = sin(y). This can be computed by a distance from the minimum (i).
        y_tmp = math.asin(xl)  # this will give me a value between -pi/2 and pi/2
        dist = y_tmp - (-pi / 2)  # this is the distance between the minimum of the sin function and a value that
        # satisfies xl = sin(y)
        lb1 = i1 + dist
        lb2 = i2 + dist
        if lb1 >= yl - feasibility_tol:
            lb = lb1
        else:
            lb = lb2
    else:
        # sin(yl) > xu
        i = (yl - pi / 2) / (2 * pi)
        i1 = math.floor(i)
        i2 = math.ceil(i)
        i1 = 2 * pi * i1 + pi / 2
        i2 = 2 * pi * i2 + pi / 2
        y_tmp = math.asin(xu)
        dist = pi / 2 - y_tmp
        lb1 = i1 + dist
        lb2 = i2 + dist
        if lb1 >= yl - feasibility_tol:
            lb = lb1
        else:
            lb = lb2

    # use the same logic for the maximum
    if yu >= inf:
        ub = yu
    elif xl <= math.sin(yu) <= xu:
        ub = yu
    elif math.sin(yu) > xu:
        i = (yu - pi / 2) / (2 * pi)
        i1 = math.ceil(i)
        i2 = math.floor(i)
        i1 = 2 * pi * i1 + pi / 2
        i2 = 2 * pi * i2 + pi / 2
        y_tmp = math.asin(xu)
        dist = pi / 2 - y_tmp
        ub1 = i1 - dist
        ub2 = i2 - dist
        if ub1 <= yu + feasibility_tol:
            ub = ub1
        else:
            ub = ub2
    else:
        # math.sin(yu) < xl
        i = (yu + pi / 2) / (2 * pi)
        i1 = math.ceil(i)
        i2 = math.floor(i)
        i1 = 2 * pi * i1 - pi / 2
        i2 = 2 * pi * i2 - pi / 2
        y_tmp = math.asin(xl)
        dist = y_tmp - (-pi / 2)
        ub1 = i1 - dist
        ub2 = i2 - dist
        if ub1 <= yu + feasibility_tol:
            ub = ub1
        else:
            ub = ub2

    return lb, ub


def acos(xl, xu, yl, yu, feasibility_tol):
    """
    y = acos(x); propagate bounds from x to y
    x = cos(y)

    Parameters
    ----------
    xl: float
    xu: float
    yl: float
    yu: float

    Returns
    -------
    lb: float
    ub: float
    """
    if xl < -1:
        xl = -1
    if xu > 1:
        xu = 1

    pi = math.pi

    if yl <= -inf:
        lb = yl
    elif xl <= math.cos(yl) <= xu:
        # if xl <= cos(yl) <= xu then yl satisfies the bounds on x, and the lower bound of y cannot be improved
        lb = yl
    elif math.cos(yl) < xl:
        """
        we can only push yl up from its current value to the next lowest value such that xl = cos(y). In other words,
        we need to

        min y
        s.t.
            xl = cos(y)
            y >= yl

        globally. 
        """
        # first find the next minimum of x = cos(y). Minimums occur at y = 2*pi*n - pi for integer n.
        i = (yl + pi) / (2 * pi)
        i1 = math.floor(i)
        i2 = math.ceil(i)
        i1 = 2 * pi * i1 - pi
        i2 = 2 * pi * i2 - pi
        # now find the next value of y such that xl = cos(y). This can be computed by a distance from the minimum (i).
        y_tmp = math.acos(xl)  # this will give me a value between 0 and pi
        dist = pi - y_tmp  # this is the distance between the minimum of the sin function and a value that
        # satisfies xl = sin(y)
        lb1 = i1 + dist
        lb2 = i2 + dist
        if lb1 >= yl - feasibility_tol:
            lb = lb1
        else:
            lb = lb2
    else:
        # cos(yl) > xu
        # first find the next maximum of x = cos(y).
        i = yl / (2 * pi)
        i1 = math.floor(i)
        i2 = math.ceil(i)
        i1 = 2 * pi * i1
        i2 = 2 * pi * i2
        y_tmp = math.acos(xu)
        dist = y_tmp
        lb1 = i1 + dist
        lb2 = i2 + dist
        if lb1 >= yl - feasibility_tol:
            lb = lb1
        else:
            lb = lb2

    # use the same logic for the maximum
    if yu >= inf:
        ub = yu
    elif xl <= math.cos(yu) <= xu:
        ub = yu
    elif math.cos(yu) > xu:
        i = yu / (2 * pi)
        i1 = math.ceil(i)
        i2 = math.floor(i)
        i1 = 2 * pi * i1
        i2 = 2 * pi * i2
        y_tmp = math.acos(xu)
        dist = y_tmp
        ub1 = i1 - dist
        ub2 = i2 - dist
        if ub1 <= yu + feasibility_tol:
            ub = ub1
        else:
            ub = ub2
    else:
        # math.cos(yu) < xl
        i = (yu + pi) / (2 * pi)
        i1 = math.ceil(i)
        i2 = math.floor(i)
        i1 = 2 * pi * i1 - pi
        i2 = 2 * pi * i2 - pi
        y_tmp = math.acos(xl)
        dist = pi - y_tmp
        ub1 = i1 - dist
        ub2 = i2 - dist
        if ub1 <= yu + feasibility_tol:
            ub = ub1
        else:
            ub = ub2

    return lb, ub


def atan(xl, xu, yl, yu):
    """
    y = atan(x); propagate bounds from x to y
    x = tan(y)

    Parameters
    ----------
    xl: float
    xu: float
    yl: float
    yu: float

    Returns
    -------
    lb: float
    ub: float
    """

    pi = math.pi

    # tan goes to -inf and inf at every pi*i + pi/2 (integer i).
    if xl <= -inf or yl <= -inf:
        lb = yl
    else:
        i = (yl - pi / 2) / pi
        i = math.floor(i)
        i = pi * i + pi / 2
        y_tmp = math.atan(xl)
        dist = y_tmp - (-pi/2)
        lb = i + dist

    if xu >= inf or yu >= inf:
        ub = yu
    else:
        i = (yu - pi / 2) / pi
        i = math.ceil(i)
        i = pi * i + pi / 2
        y_tmp = math.atan(xu)
        dist = pi / 2 - y_tmp
        ub = i - dist

    if yl > lb:
        lb = yl
    if yu < ub:
        ub = yu

    return lb, ub
